import torchvision.transforms
import torch
from PIL import Image
from typing import List
from ultralytics import YOLO
from torch.utils.data import Dataset
import torch.nn.functional as F
import cv2 as cv
import numpy as np

class MultiTransform():
    """
    MultiTransform is an extension to keep the same transformation over a sequence of images instead of initializing a new transformation for every single image. It is inspired by `torchvision.transforms` and could be used for video augmentation. Use `rsp.ml.multi_transforms.Compose`to combine multiple image sequence transformations.
    
    > **Note** `rsp.ml.multi_transforms.MultiTransform` is a base class and should be inherited.
    """
    def __init__(self):
        """
        Initializes a new instance.
        """
        pass

    def __call__(self, input):
        """
        Call self as a function.

        Parameters
        -----------
        input : torch.Tensor<br>List[PIL.Image]<br>List[numpy.array]
            Sequence of images
        """
        raise NotImplementedError()
    
    def __get_size__(self, imgs):
        if not hasattr(self, 'size'):
            if isinstance(imgs[0], torch.Tensor):
                self.size = (imgs[0].shape[1], imgs[0].shape[2])
            else:
                self.size = (imgs[0].size[1], imgs[0].size[0])
    
    def __reset__(self):
        raise NotImplementedError()

#__example__ import rsp.ml.multi_transforms as t\n
#__example__ transforms = t.Compose([
#__example__ \tt.BGR2GRAY(),
#__example__ \tt.Scale(0.5)
#__example__ ])
class Compose():
    """
    Composes several MultiTransforms together.
    """
    def __init__(self, children:List[MultiTransform]):
        """
        Initializes a new instance.

        Parameters
        ----------
        children : List[MultiTransform]
            List of MultiTransforms to compose.
        """
        self.children = children
        pass

    def __call__(self, input):
        result = input
        for c in self.children:
            result = c(result)
        for c in self.children:
            c.__reset__()
        return result
    
    def __reset__(self):
        pass

class Normalize(MultiTransform):
    """
    Normalize images with mean and standard deviation. Given mean: (mean[1],...,mean[n]) and std: (std[1],..,std[n]) for n channels, this transform will normalize each channel of the input torch.*Tensor i.e., output[channel] = (input[channel] - mean[channel]) / std[channel]
    
    > Based on torchvision.transforms.Normalize
    """
    def __init__(self, mean, std, inplace = False):
        """
        Initializes a new instance.

        Parameters
        ----------
        mean : List[float]
            Sequence of means for each channel.
        std : List[float]
            Sequence of standard deviations for each channel.
        inplace : bool
            Set to True make this operation in-place.
        """
        super().__init__()

        assert len(mean) == len(std), f'Expected mean and std to have the same dimension, but got len(mean) = {len(mean)} and len(std) = {len(std)}'

        self.normalize = torchvision.transforms.Normalize(mean, std, inplace)
        self.__toTensor__ = ToTensor()
        self.__toPILImage__ = ToPILImage()
        self.__reset__()

    def __call__(self, inputs):
        is_tensor = isinstance(inputs[0], torch.Tensor)
        if not is_tensor:
            inputs = self.__toTensor__(inputs)
            inputs = torch.stack(inputs)

        input_channels = inputs.shape[1]
        transform_channels = len(self.normalize.mean)

        assert input_channels == transform_channels, f'Expected input channels == transform channels, but got input channels = {input_channels} and len(mean) = {len(self.normalize.mean)}'

        results = []
        for res in self.normalize(inputs):
            results.append(res)

        if not is_tensor:
            results = self.__toPILImage__(results)
        return results
    
    def __reset__(self):
        pass

class RemoveBackgroundAI(MultiTransform):
    def __init__(
            self,
            p:float = 1.,
            removed_color = (0, 0, 0),
            target_classes_yolo = [0],
            yolo_model = "yolo11m-seg.pt"
    ):
        super().__init__()

        self.p = p
        self.removed_color = removed_color
        self.target_classes_yolo = target_classes_yolo

        self.__toTensor__ = ToTensor()
        self.__toPILImage__ = ToPILImage()
        self.__toCVImage__ = ToCVImage()
        self.__segmentation_model__ = YOLO(yolo_model, verbose=False)

        self.__reset__()

    def __remove_background__(self, img, bg, mask):
        """
        Changes the background of the input image.

        Parameters
        ----------
        img : np.array
            Input image
        bg : np.array
            Background image
        mask : np.array
            Mask
        """
        w, h = img.shape[1], img.shape[0]
        bg_w, bg_h = bg.shape[1], bg.shape[0]
        scale = np.min([w / bg_w, h / bg_h])
        new_w, new_h = int(np.round(scale * bg_w)), int(np.round(scale * bg_h))

        bg = cv.resize(bg, (new_w, new_h))

        img[mask > 0] = bg[mask > 0]

        return img
    
    def __get_segmentation_mask__(self, img):
        results = self.__segmentation_model__(img, conf=0.3, iou=0.45, verbose=False)
        segmentation_mask = np.zeros((img.shape[0], img.shape[1]))
        if results and results[0].masks:
            masks = results[0].masks.data.cpu().numpy()
            class_ids = results[0].boxes.cls.cpu().numpy().astype(int) 

            for class_id, mask in zip(class_ids, masks):
                if class_id not in self.target_classes_yolo:
                    continue
                mask = cv.resize(mask, (img.shape[1], img.shape[0]))
                segmentation_mask = np.logical_or(segmentation_mask, mask)
 
        segmentation_mask = np.stack([segmentation_mask] * 3, axis=-1)
        return segmentation_mask == False

    def __call__(self, inputs):
        self.__get_size__(inputs)
        self.__reset__()

        if self.__should_remove_background__:
            is_tensor = isinstance(inputs[0], torch.Tensor)
            if not is_tensor:
                inputs = self.__toTensor__(inputs)

            is_color_image = inputs[0].shape[0] == 3
            is_depth_image = inputs[0].shape[0] == 4

            if is_color_image:
                self.__masks__ = []

            # backgrounds
            bg_color = np.zeros((self.size[0], self.size[1], 3), dtype=np.float32)
            bg_color[:, :, 0] = self.removed_color[0] / 255
            bg_color[:, :, 1] = self.removed_color[1] / 255
            bg_color[:, :, 2] = self.removed_color[2] / 255

            bg = bg_color

            if is_depth_image:
                bg_depth = np.zeros((self.size[0], self.size[1], 1), dtype=np.float32)
                bg_depth = np.expand_dims(bg_depth, 2)
                bg_depth = np.concatenate([bg_depth, bg_depth, bg_depth], axis = 2)

                bg_depth = np.expand_dims(bg_depth, 2)

            results = []
            for i, input in enumerate(self.__toCVImage__(inputs)):
                img = np.asarray(input * 255, dtype=np.uint8)
                img_rgb = img[:, :, 0:3]

                mask = self.__get_segmentation_mask__(img_rgb)
                #frame[mask == 0] = 0

                mask = np.asarray(mask, dtype=np.uint8)
                
                if is_depth_image:
                    bg = np.concatenate([bg_color, bg_depth], axis = 2)

                result = self.__remove_background__(input, bg, mask)

                results.append(result)

            results = self.__toTensor__(results)

            if not is_tensor:
                results = self.__toPILImage__(results)
        else:
            results = inputs
        return results

    def __reset__(self):
        self.__should_remove_background__ = np.random.random() < self.p

#__example__ from rsp.nl.dataset import TUCRID
#__example__ import rsp.ml.multi_transforms as multi_transforms
#__example__ 
#__example__ USE_DEPTH_DATA = False
#__example__ backgrounds = TUCRID.load_backgrounds(USE_DEPTH_DATA)
#__example__ tranforms_train = multi_transforms.Compose([
#__example__     multi_transforms.ReplaceBackground(
#__example__         backgrounds = backgrounds,
#__example__         hsv_filter=[(69, 87, 139, 255, 52, 255)],
#__example__         p = 0.8
#__example__     ),
#__example__     multi_transforms.Stack()
#__example__ ])
#__example__ tucrid = TUCRID('train', load_depth_data=USE_DEPTH_DATA, transforms=tranforms_train)
#__example__ 
#__example__ for X, T in tucrid:
#__example__     for x in X:
#__example__         img = x.permute(1, 2, 0).numpy()
#__example__ 
#__example__         cv.imshow('img', img)
#__example__         cv.waitKey(30)
class ReplaceBackground(MultiTransform):
    """
        Transformation for background replacement based on HSV values. Supports depth background replacement. backgrounds have to be passed as list of tuples of rgb and depth images.
    """
    def __init__(
            self,
            backgrounds:List[np.array],
            hsv_filter:List[tuple[int, int, int, int, int, int]] = [(69, 87, 139, 255, 52, 255)],
            p:float = 1.,
            rotate:float = 5,
            max_scale:float = 2,
            max_noise:float = 0.002
        ):
        """
        Transformation for background replacement based on HSV values. Supports depth background replacement. backgrounds have to be passed as list of tuples of rgb and depth images.

        Parameters
        ----------
        backgrounds : List[np.array]
            List of background images
        hsv_filter : List[tuple[int, int, int, int, int, int]]
            List of HSV filters
        p : float, default = 1.
            Probability of applying the transformation
        rotate : float, default = 5
            Maximum rotation angle
        max_scale : float, default = 2
            Maximum scaling factor
        max_noise : float, default = 0.002
            Maximum noise level
        """
        super().__init__()
        self.backgrounds = backgrounds
        self.hsv_filter = hsv_filter
        self.p = p

        self.__toTensor__ = ToTensor()
        self.__toPILImage__ = ToPILImage()
        self.__toCVImage__ = ToCVImage()

        self.transforms:List[MultiTransform] = [
            ToTensor(),
            Rotate(rotate),
            RandomCrop(max_scale = max_scale),
            RandomHorizontalFlip(),
            RandomVerticalFlip(),
            GaussianNoise(max_noise_level=max_noise),
            ToCVImage()
        ]

    def __hsv_filter__(self, img, hmin, hmax, smin, smax, vmin, vmax, inverted):
        """
        Filters the input image based on HSV values.

        Parameters
        ----------
        img : np.array
            Input image
        hmin : int
            Minimum hue value
        hmax : int
            Maximum hue value
        smin : int
            Minimum saturation value
        smax : int
            Maximum saturation value
        vmin : int
            Minimum value value
        vmax : int
            Maximum value value
        inverted : bool
            Invert the mask
        """
        hsv = cv.cvtColor(img, cv.COLOR_BGR2HSV)
        lower = (hmin, smin, vmin)
        upper = (hmax, smax, vmax)
        mask = cv.inRange(hsv, lower, upper)
        mask = cv.dilate(mask, np.ones((3, 3), np.uint8), iterations=1)

        if inverted:
            mask = 255 - mask
        return mask

    def __change_background__(self, img, bg, mask):
        """
        Changes the background of the input image.

        Parameters
        ----------
        img : np.array
            Input image
        bg : np.array
            Background image
        mask : np.array
            Mask
        """
        w, h = img.shape[1], img.shape[0]
        bg_w, bg_h = bg.shape[1], bg.shape[0]
        scale = np.min([w / bg_w, h / bg_h])
        new_w, new_h = int(np.round(scale * bg_w)), int(np.round(scale * bg_h))

        bg = cv.resize(bg, (new_w, new_h))

        img[mask > 0] = bg[mask > 0]

        return img

    def __call__(self, inputs):
        self.__get_size__(inputs)
        self.__reset__()

        if self.__replace_background__:
            is_tensor = isinstance(inputs[0], torch.Tensor)
            if not is_tensor:
                inputs = self.__toTensor__(inputs)

            is_color_image = inputs[0].shape[0] == 3
            is_depth_image = inputs[0].shape[0] == 4

            if is_color_image:
                self.__masks__ = []

            # backgrounds
            bg_color = np.asarray(self.__background__[0] / 255, dtype=np.float32)

            bg_color = [bg_color]
            for t in self.transforms:
                bg_color = t(bg_color)
            bg_color = bg_color[0]
            bg = bg_color

            if is_depth_image and self.__background__[1]:
                bg_depth = np.asarray(self.__background__[1] / 255, dtype=np.float32)
                bg_depth = np.expand_dims(bg_depth, 2)
                bg_depth = np.concatenate([bg_depth, bg_depth, bg_depth], axis = 2)

                bg_depth = [bg_depth]
                for t in self.transforms:
                    bg_depth = t(bg_depth)
                bg_depth = bg_depth[0][:,:, 0]
                bg_depth = np.expand_dims(bg_depth, 2)

            results = []
            for i, input in enumerate(self.__toCVImage__(inputs)):
                img = np.asarray(input * 255, dtype=np.uint8)
                img_rgb = img[:, :, 0:3]

                mask = np.ones((input.shape[0], input.shape[1]))
                for f in self.hsv_filter:
                    hsv_mask = self.__hsv_filter__(img_rgb.copy(), f[0], f[1], f[2], f[3], f[4], f[5], inverted = False)
                    hsv_mask = hsv_mask / 255

                    mask = cv.bitwise_and(mask, hsv_mask)

                mask = np.asarray(mask, dtype=np.uint8)
                
                if is_depth_image:
                    if self.__background__[1]:
                        bg = np.concatenate([bg_color, bg_depth], axis = 2)
                    else:
                        bg = np.concatenate([bg_color, input[:, :, 3:4]], axis = 2)

                result = self.__change_background__(input, bg, mask)

                results.append(result)

            results = self.__toTensor__(results)

            for t in self.transforms:
                t.__reset__()

            if not is_tensor:
                results = self.__toPILImage__(results)
        else:
            results = inputs
        return results

    def __reset__(self):
        self.__replace_background__ = np.random.random() < self.p
        idx = np.random.randint(0, len(self.backgrounds))

        if isinstance(self.backgrounds, Dataset):
            self.__background__ = self.backgrounds[idx][0]  # avoid sampling target
        else:
            self.__background__ = self.backgrounds[idx]     # sample from list of backgrounds

        if not isinstance(self.__background__, tuple):
            self.__background__ = (self.__background__, None)

class ToTensor(MultiTransform):
    """
    Converts a sequence of images to torch.Tensor.
    """
    def __init__(self):
        super().__init__()

        self.toTensor = torchvision.transforms.ToTensor()
        self.__reset__()

    def __call__(self, images) -> List[torch.Tensor]:
        results = []
        for img in images:
            result = self.toTensor(img).float()
            results.append(result)
        return results
    
    def __reset__(self):
        pass

#__image__ ![](documentation/image/multi_transforms.CenterCrop.png)
class CenterCrop(MultiTransform):
    """
    Crops Images at the center after upscaling them. Dimensions kept the same.
    """
    def __init__(self, max_scale = 2):
        """
        Initializes a new instance.

        Parameters
        ----------
        max_scale : float
            Images are scaled randomly between 1. and max_scale before cropping to original size.
        """
        super().__init__()

        if max_scale < 1:
            raise Exception(f'max_scale expected to be greater than 1. Actual value is {max_scale})')
        self.max_scale = max_scale

        self.__toTensor__ = ToTensor()
        self.__toPILImage__ = ToPILImage()

        self.__reset__()

    def __call__(self, imgs):
        self.__get_size__(imgs)
    
        results = []

        is_tensor = isinstance(imgs[0], torch.Tensor)
        if not is_tensor:
            imgs = self.__toTensor__(imgs)
        
        for img in imgs:
            img_before = img.permute(1, 2, 0).numpy()

            w, h = self.size[1], self.size[0]
            new_w, new_h = int(np.round(w * self.__scale__)), int(np.round(h * self.__scale__))
            img_after = cv.resize(img_before, (new_w, new_h))

            cx, cy = new_w // 2, new_h // 2
            result = img_after[cy - h // 2: cy + h // 2, cx - w // 2: cx + w // 2]
            result = torch.tensor(result, dtype=img.dtype).permute(2, 0, 1)

            results.append(result)

        if not is_tensor:
            results = self.__toPILImage__(results)
            
        return results
    
    def __reset__(self):
        self.__scale__ = 1. + np.random.random() * (self.max_scale - 1.)

#__image__ ![](documentation/image/multi_transforms.RandomCrop.png)
class RandomCrop(MultiTransform):
    """
    Crops Images at a random location after upscaling them. Dimensions kept the same.
    """
    def __init__(self, max_scale = 2):
        """
        Initializes a new instance.

        Parameters
        ----------
        max_scale : float
            Images are scaled randomly between 1. and max_scale before cropping to original size.
        """
        super().__init__()

        if max_scale < 1:
            raise Exception(f'max_scale expected to be greater than 1. Actual value is {max_scale})')
        self.max_scale = max_scale

        self.__toCVImage__ = ToCVImage()
        self.__toTensor__ = ToTensor()
        self.__toPILImage__ = ToPILImage()

    def __call__(self, imgs):
        self.__get_size__(imgs)
        if not hasattr(self, '__scale__'):
            self.__reset__()
    
        results = []

        is_tensor = isinstance(imgs[0], torch.Tensor)
        if not is_tensor:
            imgs = self.__toTensor__(imgs)
        
        #imgs = self.__toCVImage__(imgs)

        for img in imgs:
            img_before = img.permute(1, 2, 0).numpy()

            w, h = self.size[1], self.size[0]
            new_w, new_h = int(np.round(w * self.__scale__)), int(np.round(h * self.__scale__))
            img_after = cv.resize(img_before, (new_w, new_h))

            img_after = torch.tensor(img_after, dtype=img.dtype).permute(2, 0, 1)
            result = img_after[:, self.__sy__:self.__sy__ + h, self.__sx__:self.__sx__ + w]

            results.append(result)

        if not is_tensor:
            results = self.__toPILImage__(results)
            
        return results
    
    def __reset__(self):
        self.__scale__ = 1. + np.random.random() * (self.max_scale - 1.)
        if not hasattr(self, 'size'):
            self.__sx__ = 0
            self.__sy__ = 0
        else:
            w, h = self.size[1], self.size[0]
            new_w, new_h = self.__scale__ * self.size[1], self.__scale__ * self.size[0]
            self.__sx__ = int(np.round(np.random.random() * (new_w - w)))
            self.__sy__ = int(np.round(np.random.random () * (new_h - h)))

#__image__ ![](documentation/image/multi_transforms.Rotate.png)
#__equation__ $angle = -max\_angle + 2 \cdot random() \cdot max\_angle$
class Rotate(MultiTransform):
    """
    Randomly rotates images.
    """
    def __init__(self, max_angle = 180, auto_scale:bool = True):
        """
        Iitializes a new instance.

        Parameters
        ----------
        max_angle : float
            Maximal rotation in degrees | -max_angle <= rotate <= max_angle
        auto_scale : bool, default = True
            Image will be resized when auto scale is activated to avoid black margins.
        """
        super().__init__()

        self.max_angle = max_angle
        self.auto_scale = auto_scale

        self.__toTensor__ = ToTensor()
        self.__toPILImage__ = ToPILImage()

    def __call__(self, imgs):
        self.__get_size__(imgs)
        if not hasattr(self, '__angle__'):
            self.__reset__()
    
        results = []

        is_tensor = isinstance(imgs[0], torch.Tensor)
        if not is_tensor:
            imgs = self.__toTensor__(imgs)

        for img in imgs:
            img_before = img.permute(1, 2, 0).numpy()

            w, h = self.size[1], self.size[0]
            mat = cv.getRotationMatrix2D((w // 2, h // 2), self.__angle__, self.__scale__)
            img_after = cv.warpAffine(img_before, mat, (w, h))

            if len(img_after.shape) == 2:   # grayscale image
                img_after = np.expand_dims(img_after, 2)

            result = torch.tensor(img_after, dtype=img.dtype).permute(2, 0, 1)

            results.append(result)

        if not is_tensor:
            results = self.__toPILImage__(results)
            
        return results
    
    def __reset__(self):
        self.__angle__ = -self.max_angle + 2 * np.random.random() * self.max_angle

        w, h = self.size[1], self.size[0]
        new_w = w + np.abs(np.sin(self.__angle__ / 180 * np.pi) * w)
        new_h = h + np.abs(np.sin(self.__angle__ / 180 * np.pi) * h)

        self.__scale__ = 1.03 * np.max([new_w / w, new_h / h]) if self.auto_scale else 1.

class ToNumpy(MultiTransform):
    """
    Converts a `torch.Tensor`to `numpy`
    """
    def __init__(self):
        super().__init__()

        self.__reset__()

    def __call__(self, tensor:torch.Tensor):
        result = tensor.numpy()
        return result
    
    def __reset__(self):
        pass

class ToCVImage(MultiTransform):
    """
    Converts a `torch.Tensor`to Open CV image by changing dimensions (d0, d1, d2) -> (d1, d2, d0) and converting `torch.Tensor` to `numpy`.
    """
    def __init__(self):
        super().__init__()

        self.__toTensor__ = ToTensor()

        self.__reset__()

    def __call__(self, inputs) -> List[np.array]:
        is_tensor = isinstance(inputs[0], torch.Tensor)
        if not is_tensor:
            inputs = self.__toTensor__(inputs)
        results = []
        for img in inputs:
            result = np.asarray(img.permute(1, 2, 0), dtype = np.float32)
            results.append(result)
        return results
    
    def __reset__(self):
        pass

class ToPILImage(MultiTransform):
    """
    Converts sequence of images to sequence of `PIL.Image`.
    """
    def __init__(self):
        super().__init__()

        self.__toPILImage__ = torchvision.transforms.ToPILImage()

        self.__reset__()

    def __call__(self, tensor:torch.Tensor):
        results = []
        for img in tensor:
            result = self.__toPILImage__(img)
            results.append(result)
        return results
    
    def __reset__(self):
        pass

class BGR2RGB(MultiTransform):
    """
    Converts sequence of BGR images to RGB images.
    """
    def __init__(self):
        super().__init__()

        self.__toTensor__ = ToTensor()
        self.__toPILImage__ = ToPILImage()

        self.__reset__()

    def __call__(self, inputs):
        is_tensor = isinstance(inputs[0], torch.Tensor)
        if not is_tensor:
            inputs = self.__toTensor__(inputs)

        results = []
        for input in inputs:
            result = input.clone()
            result[0:3] = torch.flip(input[0:3], (0,))
            results.append(result)

        if not is_tensor:
            results = self.__toPILImage__(results)
        return results
    
    def __reset__(self):
        pass

class RGB2BGR(BGR2RGB):
    """
    Converts sequence of RGB images to BGR images.
    """
    pass

class Scale(MultiTransform):
    def __init__(self, scale):
        super().__init__()

        self.__toTensor__ = ToTensor()
        self.__toPILImage__ = ToPILImage()

        self.scale = scale

        self.__toCVImage__ = ToCVImage()

        self.__reset__()

    def __call__(self, inputs):
        self.__get_size__(inputs)

        is_tensor = isinstance(inputs[0], torch.Tensor)
        if not is_tensor:
            inputs = self.__toTensor__(inputs)

        results = []
        for input in self.__toCVImage__(inputs):
            w, h = self.size[1], self.size[0]
            new_w, new_h = int(np.round(self.scale * w)), int(np.round(self.scale * h))
            result = cv.resize(input, (new_w, new_h))
            results.append(result)

        results = self.__toTensor__(results)

        if not is_tensor:
            results = self.__toPILImage__(results)
        return results
    
    def __reset__(self):
        pass

class Resize(MultiTransform):
    def __init__(self, target_size:tuple[int, int], auto_crop:bool = True):
        super().__init__()

        self.__toTensor__ = ToTensor()
        self.__toPILImage__ = ToPILImage()
        self.__toCVImage__ = ToCVImage()

        self.target_size = target_size
        self.auto_crop = auto_crop

        self.__reset__()

    def __call__(self, inputs):
        self.__get_size__(inputs)

        is_tensor = isinstance(inputs[0], torch.Tensor)
        if not is_tensor:
            inputs = self.__toTensor__(inputs)

        results = []
        for input in self.__toCVImage__(inputs):
            w, h = self.size[1], self.size[0]
            if self.auto_crop:
                scale = np.max([self.target_size[0] / self.size[0], self.target_size[1] / self.size[1]])
                new_w, new_h = int(np.round(scale * w)), int(np.round(scale * h))
                result = cv.resize(input, (new_w, new_h))

                cx, cy = result.shape[1] // 2, result.shape[0] // 2
                result = result[cy-self.target_size[0]//2:cy+self.target_size[0]//2, cx-self.target_size[1]//2:cx+self.target_size[1]//2]
            else:
                new_w, new_h = self.target_size[1], self.target_size[0]
                result = cv.resize(input, (new_w, new_h))
            results.append(result)

        results = self.__toTensor__(results)

        if not is_tensor:
            results = self.__toPILImage__(results)
        return results
    
    def __reset__(self):
        pass

class Brightness(MultiTransform):
    def __init__(self, min_rel:float, max_rel:float):
        super().__init__()

        if min_rel < 0 or max_rel < 0 or min_rel > max_rel:
            raise Exception(f'min_rel and max_rel expected to be greater or equal 0. min_rel expected to be less or equal max_rel')

        self.__toTensor__ = ToTensor()
        self.__toPILImage__ = ToPILImage()

        self.min_rel = min_rel
        self.max_rel = max_rel

        self.__toCVImage__ = ToCVImage()

        self.__reset__()

    def __call__(self, inputs):
        assert inputs[0].shape[2] >= 3, f'Expected input channels >= 3 but got input[0].shape = {input[0].shape}'

        self.__get_size__(inputs)
        is_tensor = isinstance(inputs[0], torch.Tensor)
        if not is_tensor:
            inputs = self.__toTensor__(inputs)

        results = []
        for input in self.__toCVImage__(inputs):
            hsv = cv.cvtColor(input, cv.COLOR_BGR2HSV)
            h, s, v = cv.split(hsv)

            v *= self.rel
            v[v > 1] = 1
            v[v < 0] = 0
            hsv = cv.merge((h, s, v))
            result = np.copy(input)
            result[:, :, 0:3] = cv.cvtColor(hsv, cv.COLOR_HSV2BGR)

            results.append(result)
        
        results = self.__toTensor__(results)
        
        if not is_tensor:
            results = self.__toPILImage__(results)
        return results
    
    def __reset__(self):
        self.rel = self.min_rel + np.random.random() * (self.max_rel - self.min_rel)

class Satturation(MultiTransform):
    def __init__(self, min_rel:float, max_rel:float):
        super().__init__()

        if min_rel < 0 or max_rel < 0 or min_rel > max_rel:
            raise Exception(f'min_rel and max_rel expected to be greater or equal 0. min_rel expected to be less or equal max_rel')

        self.__toTensor__ = ToTensor()
        self.__toPILImage__ = ToPILImage()

        self.min_rel = min_rel
        self.max_rel = max_rel

        self.__toCVImage__ = ToCVImage()

        self.__reset__()

    def __call__(self, inputs):
        assert inputs[0].shape[2] >= 3, f'Expected input channels >= 3 but got input[0].shape = {input[0].shape}'

        self.__get_size__(inputs)
        is_tensor = isinstance(inputs[0], torch.Tensor)
        if not is_tensor:
            inputs = self.__toTensor__(inputs)

        results = []
        for input in self.__toCVImage__(inputs):
            hsv = cv.cvtColor(input, cv.COLOR_BGR2HSV)
            h, s, v = cv.split(hsv)

            s *= self.rel
            s[s > 1] = 1
            s[s < 0] = 0
            hsv = cv.merge((h, s, v))
            result = np.copy(input)
            result[:, :, 0:3] = cv.cvtColor(hsv, cv.COLOR_HSV2BGR)

            results.append(result)

        results = self.__toTensor__(results)

        if not is_tensor:
            results = self.__toPILImage__(results)
        return results
    
    def __reset__(self):
        self.rel = self.min_rel + np.random.random() * (self.max_rel - self.min_rel)

class Color(MultiTransform):
    def __init__(self, max_rel:float, p = 1.):
        super().__init__()

        if max_rel < 0. or max_rel > 1.:
            raise Exception(f'Expected 0 <= max_rel <= 1, but got max_rel = {max_rel}')

        assert p >= 0. and p <= 1., f'Expected 0 <= p <= 1, but got p = {p}'
        self.p = p

        self.__toTensor__ = ToTensor()
        self.__toPILImage__ = ToPILImage()

        self.max_rel = max_rel

        self.__toCVImage__ = ToCVImage()

        self.__reset__()

    def __call__(self, inputs):
        is_tensor = isinstance(inputs[0], torch.Tensor)
        if is_tensor:
            assert inputs[0].shape[2] >= 3, f'Expected input channels >= 3 but got input[0].shape = {input[0].shape}'
        elif isinstance(inputs[0], Image.Image):
            assert inputs[0].mode in ['RGB', 'RGBA', 'CMYK'], f'Expected input channels >= 3 but got input[0].shape = {input[0].shape}'
        
        self.__get_size__(inputs)
        
        if not is_tensor:
            inputs = self.__toTensor__(inputs)
        
        results = []
        for input in self.__toCVImage__(inputs):
            hsv = cv.cvtColor(input, cv.COLOR_BGR2HSV)
            h, s, v = cv.split(hsv)

            h += self.offset_h
            h[h > 360] = h[h > 360] - 360
            hsv = cv.merge((h, s, v))
            result = np.copy(input)
            result[:, :, 0:3] = cv.cvtColor(hsv, cv.COLOR_HSV2BGR)

            results.append(result)
        
        results = self.__toTensor__(results)

        if not is_tensor:
            results = self.__toPILImage__(results)
        return results
    
    def __reset__(self):
        if np.random.random() < self.p:
            rel = -self.max_rel + 2 * np.random.random() * self.max_rel
            self.offset_h = rel * 360
        else:
            self.offset_h = 0

class GaussianNoise(MultiTransform):
    def __init__(self, min_noise_level = 0., max_noise_level:float = 0.005):
        super().__init__()

        self.__toTensor__ = ToTensor()
        self.__toPILImage__ = ToPILImage()

        self.min_noise_level = min_noise_level
        self.max_noise_level = max_noise_level

        self.__toCVImage__ = ToCVImage()

        self.__reset__()

    def __call__(self, inputs):
        self.__get_size__(inputs)
        is_tensor = isinstance(inputs[0], torch.Tensor)
        if not is_tensor:
            inputs = self.__toTensor__(inputs)

        results = []
        for input in self.__toCVImage__(inputs):
            noise = -self.__noise_level__ + 2 * np.random.random(input.shape) * self.__noise_level__
            result = input + noise

            results.append(result)

        results = self.__toTensor__(results)

        if not is_tensor:
            results = self.__toPILImage__(results)
        return results
    
    def __reset__(self):
        self.__noise_level__ = self.min_noise_level + np.random.random() * (self.max_noise_level - self.min_noise_level)

class Stack(MultiTransform):
    def __init__(self):
        super().__init__()

        self.__toTensor__ = ToTensor()
        self.__toPILImage__ = ToPILImage()

        self.__reset__()

    def __call__(self, inputs):
        self.__get_size__(inputs)
        is_tensor = isinstance(inputs[0], torch.Tensor)
        if not is_tensor:
            inputs = self.__toTensor__(inputs)

        results = torch.stack(inputs)
        
        return results
    
    def __reset__(self):
        pass

class BGR2GRAY(MultiTransform):
    """
    Converts a sequence of BGR images to grayscale images.
    """
    def __init__(self):
        super().__init__()

        self.__toTensor__ = ToTensor()
        self.__toCVImage__ = ToCVImage()
        self.__toPILImage__ = ToPILImage()

        self.__reset__()

    def __call__(self, inputs):
        self.__get_size__(inputs)
        is_tensor = isinstance(inputs[0], torch.Tensor)
        if not is_tensor:
            inputs = self.__toTensor__(inputs)

        results = []
        for input in self.__toCVImage__(inputs):
            result = cv.cvtColor(input, cv.COLOR_BGR2GRAY)

            results.append(result)
        
        results = self.__toTensor__(results)
        
        if not is_tensor:
            results = self.__toPILImage__(results)
        return results
    
    def __reset__(self):
        pass

class RandomHorizontalFlip(MultiTransform):
    def __init__(self):
        super().__init__()

        self.__toTensor__ = ToTensor()
        self.__toCVImage__ = ToCVImage()
        self.__toPILImage__ = ToPILImage()

        self.__reset__()

    def __call__(self, inputs):
        self.__get_size__(inputs)
        is_tensor = isinstance(inputs[0], torch.Tensor)
        if not is_tensor:
            inputs = self.__toTensor__(inputs)

        results = []
        for input in self.__toCVImage__(inputs):
            if self.__should_flip__:
                result = cv.flip(input, 1)
            else:
                result = input

            results.append(result)
        
        results = self.__toTensor__(results)
        
        if not is_tensor:
            results = self.__toPILImage__(results)
        return results
    
    def __reset__(self):
        self.__should_flip__ = np.random.random() > 0.5

class RandomVerticalFlip(MultiTransform):
    def __init__(self):
        super().__init__()

        self.__toTensor__ = ToTensor()
        self.__toCVImage__ = ToCVImage()
        self.__toPILImage__ = ToPILImage()

        self.__reset__()

    def __call__(self, inputs):
        self.__get_size__(inputs)
        is_tensor = isinstance(inputs[0], torch.Tensor)
        if not is_tensor:
            inputs = self.__toTensor__(inputs)

        results = []
        for input in self.__toCVImage__(inputs):
            if self.__should_flip__:
                result = cv.flip(input, 0)
            else:
                result = input

            results.append(result)
        
        results = self.__toTensor__(results)
        
        if not is_tensor:
            results = self.__toPILImage__(results)
        return results
    
    def __reset__(self):
        self.__should_flip__ = np.random.random() > 0.5

if __name__ == '__main__':
    from rsp.ml.dataset import TUCRID
    
    USE_DEPTH_DATA = True

    backgrounds = TUCRID.load_backgrounds(USE_DEPTH_DATA)
    tranforms_train = Compose([
        ReplaceBackground(
            backgrounds = backgrounds,
            hsv_filter=[(69, 87, 139, 255, 52, 255)],
            p = 0.8,
            rotate=0,
            max_scale=2,
            max_noise=0.002
        ),
        Resize((400, 400), auto_crop=False),
        Color(0.1, p = 0.2),
        Brightness(0.7, 1.3),
        Satturation(0.7, 1.3),
        RandomHorizontalFlip(),
        GaussianNoise(0.002),
        Rotate(max_angle=3),
        Stack()
    ])
    tucreid = TUCRID('train', load_depth_data=USE_DEPTH_DATA, transforms=tranforms_train)

    for X, T in tucreid:
        for x in X:
            img_color = x[0:3].permute(1, 2, 0).numpy()
            img_depth = np.expand_dims(x[3].numpy(), 2)
            #img_depth = np.concatenate([img_depth, img_depth, img_depth], axis = 2)

            img_color = np.array(img_color * 255, dtype=np.uint8)
            img_depth = np.array(img_depth * 255, dtype=np.uint8)

            img_depth = cv.applyColorMap(img_depth, cv.COLORMAP_JET)

            img = np.hstack([img_color, img_depth])

            cv.imshow('img', img)
            cv.waitKey()
    
    pass


from datetime import datetime
from torch.utils.data import DataLoader
from glob import glob
from pathlib import Path
from tqdm import tqdm
import numpy as np
import json
import matplotlib.pyplot as plt
import pickle as pkl
import copy
import torch
import torch.nn as nn
import pandas as pd

try:
    import rsp.common.console as console
except Exception as e:
    print(e)

#__example__ from rsp.ml.run import Run
#__example__ import rsp.ml.metrics as m
#__example__
#__example__ metrics = [
#__example__     m.top_1_accuracy
#__example__ ]
#__example__ config = {
#__example__     m.top_1_accuracy.__name__: {
#__example__         'ymin': 0,
#__example__         'ymax': 1
#__example__     }
#__example__ }
#__example__ run = Run(id='run0001', metrics=metrics, config=config, ignore_outliers_in_chart_scaling=True)
#__example__
#__example__ for epoch in range(100):
#__example__     """here goes some training code, giving us inputs, predictions and targets"""
#__example__     acc = m.top_1_accuracy(predictions, targets)
#__example__     run.append(m.top_1_accuracy.__name__, 'train', acc)
class Run():
    """
    Run class to store and manage training
    """
    def __init__(
            self,
            id = None,
            moving_average_epochs = 1,
            metrics = None,
            device:str = None,
            ignore_outliers_in_chart_scaling:bool = False,
            config:dict = {}
        ):
        """
        Run class to store and manage training

        Parameters:
        ----------
        id : str, default = None
            Id of the run. If None, a new id is generated
        moving_average_epochs : int, default = 1
            Number of epochs to average over
        metrics : list, default = None
            List of metrics to compute. Each metric should be a function that takes Y and T as input.
        device : str, default = None
            torch device to run on
        ignore_outliers_in_chart_scaling : bool, default = False
            Ignore outliers when scaling charts
        config : dict, default = {}
            Configuration dictionary. Keys are metric names and values are dictionaries with keys 'ymin' and 'ymax'
        """
        if id is None:
            self.id = datetime.now().strftime('%Y%m%d%H%M%S%f')
            self.data = {
                'start_time': datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')
            }
        else:
            self.id = id

        self.metrics = metrics
        self.ignore_outliers_in_chart_scaling = ignore_outliers_in_chart_scaling
        self.config = config
        self.config['time'] = {'ymin': 0}

        if device is None:
            if torch.cuda.is_available():
                self.device = 'cuda'
            elif torch.backends.mps.is_available():
                self.device = 'mps'
            else:
                self.device = 'cpu'
        else:
            self.device = device
            
        self.moving_average_epochs = moving_average_epochs
        self.__init_run_dir__()
        self.__load__()
    
    def append(self, key:str, phase:str, value:float):
        """
        Append value to key in phase.

        Parameters:
        ----------
        key : str
            Key to append to
        phase : str
            Phase to append to
        value : float
            Value to append
        """
        mavg_epochs = self.moving_average_epochs if self.moving_average_epochs > 0 else 1
        if not key in self.data:
            self.data[key] = {}
        if not phase in self.data[key]:
            self.data[key][phase] = {
                'val': [],
                'avg': []
            }
        if np.isnan(value):
            if len(self.data[key][phase]['val']) > 0:
                value = self.data[key][phase]['val'][-1]
            else:
                value = 0.

        self.data[key][phase]['val'].append(value)
        self.data[key][phase]['avg'].append(np.average(self.data[key][phase]['val'][-mavg_epochs:]))

    def plot(self):
        """
        Plot all keys to runs/{id}/plot/{key}.jpg
        """
        self.__init_run_dir__()

        def get_outlier_fence(data):
            df = pd.DataFrame({'data': data})
            q1 = df['data'].quantile(0.25)
            q3 = df['data'].quantile(0.75)
            iqr = q3 - q1
            fence_low = q1 - 1.2 * iqr
            fence_high = q3 + 1.2 * iqr
            return fence_low, fence_high

        for key in self.data:
            if key == 'start_time':
                continue
            key_str = key.replace('_', ' ')
            plot_file = self.directory_plot.joinpath(f'{key}.jpg')

            cmap = plt.get_cmap('tab20b')
            colors = cmap(np.linspace(0, 1, len(self.data[key])))

            ymin, ymax = 1e10, 0.
            for i, phase in enumerate(self.data[key]):
                if len(self.data[key][phase]['val']) == 0:
                    continue

                if self.ignore_outliers_in_chart_scaling:
                    val_fence = get_outlier_fence(self.data[key][phase]['val'])
                    avg_fence = get_outlier_fence(self.data[key][phase]['avg'])
                    ymin = min(ymin, min(val_fence[0], avg_fence[0]))
                    ymax = max(ymax, max(val_fence[1], avg_fence[1]))

                plt.plot(self.data[key][phase]['val'], color=colors[i], alpha=0.3 if key != 'time' else 1)
                if key != 'time':
                    plt.plot(self.data[key][phase]['avg'], label=phase, color=colors[i])

            #region chart scaling
            if not self.ignore_outliers_in_chart_scaling:
                _, _, ymin, ymax = plt.axis()

            if key in self.config:
                if 'ymin' in self.config[key]:
                    ymin = self.config[key]['ymin']
                if 'ymax' in self.config[key]:
                    ymax = self.config[key]['ymax']
            plt.ylim(ymin, ymax)
            #endregion

            plt.title(key_str)
            plt.xlabel('episode')
            if key == 'time':
                plt.ylabel(f'{key_str} [h]')
            elif key == 'time_per_sample':
                plt.ylabel(f'{key_str} [s]')
            else:
                plt.ylabel(key_str)
            
            plt.minorticks_on()
            plt.grid(which='minor', color='lightgray', linewidth=0.2)
            plt.grid(which='major', linewidth=.6)
            if key != 'time':
                plt.legend()
            plt.savefig(plot_file)
            plt.close()

    def recalculate_moving_average(self):
        """
        Recalculate moving average
        """
        mavg_epochs = self.moving_average_epochs if self.moving_average_epochs > 0 else 1
        for key in self.data:
            if key == 'start_time':
                continue
            for phase in self.data[key]:
                for i in range(len(self.data[key][phase]['val'])):
                    s_i = 0 if i + 1 - mavg_epochs < 0 else i + 1 - mavg_epochs
                    e_i = i + 1
                    test = self.data[key][phase]['val'][s_i:e_i]
                    self.data[key][phase]['avg'][i] = np.average(self.data[key][phase]['val'][s_i:e_i])
                self.data[key][phase]['avg'] = self.data[key][phase]['avg'][:len(self.data[key][phase]['val'])]

    def train_epoch(
            self,
            dataloader:DataLoader,
            model:torch.nn.Module,
            optimizer:torch.optim.Optimizer,
            criterion:torch.nn.Module,
            num_batches:int = None,
            return_YT:bool = False
        ):
        """
        Train one epoch.

        Parameters:
        ----------
        dataloader : DataLoader
            DataLoader to train on
        model : torch.nn.Module
            Model to train
        optimizer : torch.optim.Optimizer
            Optimizer to use
        criterion : torch.nn.Module
            Criterion to use
        num_batches : int, default = None
            Number of batches to train on. If None, train on all batches
        return_YT : bool, default = False
            Append Y and T to results
        
        Returns:
        -------
        results : dict
            Dictionary with results
        """
        results = self.__compute__(
            model = model,
            optimizer = optimizer,
            criterion = criterion,
            dataloader = dataloader,
            train = True,
            num_batches = num_batches,
            return_XYT = return_YT)
        for key in results:
            if key == 'X' or key == 'Y' or key == 'T':
                continue
            self.append(key, phase = 'train', value = results[key])

        return results
    
    def validate_epoch(
            self,
            dataloader:DataLoader,
            model:torch.nn.Module,
            optimizer:torch.optim.Optimizer,
            criterion:torch.nn.Module,
            num_batches:int = None,
            return_YT:bool = False
        ):
        """
        Validate one epoch.

        Parameters:
        ----------
        dataloader : DataLoader
            DataLoader to validate on
        model : torch.nn.Module
            Model to validate
        optimizer : torch.optim.Optimizer
            Optimizer to use
        criterion : torch.nn.Module
            Criterion to use
        num_batches : int, default = None
            Number of batches to validate on. If None, validate on all batches
        return_YT : bool, default = False
            Append Y and T to results

        Returns:
        -------
        results : dict
            Dictionary with results
        """
        results = self.__compute__(
            model = model,
            optimizer = optimizer,
            criterion = criterion,
            dataloader = dataloader,
            train = False,
            num_batches = num_batches,
            return_XYT = return_YT)
        for key in results:
            if key == 'X' or key == 'Y' or key == 'T':
                continue
            self.append(key, phase = 'val', value = results[key])

        return results

    def __compute__(
            self,
            model:nn.Module,
            optimizer:torch.optim.Optimizer,
            criterion:nn.Module,
            dataloader:DataLoader,
            num_batches:int,
            train:bool,
            return_XYT:bool = False
        ):
        start = datetime.now()
        iterator = iter(dataloader)
        if num_batches is None or hasattr(dataloader.dataset, '__len__') and num_batches > len(dataloader):
            num_batches = len(dataloader)
        if num_batches is None:
            num_batches = 1e10
        
        results = {
            'loss': []
        }
        if return_XYT:
            results['Y'] = []
            results['X'] = []
            results['T'] = []

        progress = tqdm(range(num_batches), desc='train' if train else 'val', total=num_batches, leave=False)
        for i in progress:
            if i >= num_batches:
                break
            
            try:
                X, T = next(iterator)
            except:
                break
            X:torch.Tensor = X.to(self.device)
            T:torch.Tensor = T.to(self.device)

            if train:
                optimizer.zero_grad()
                model.train()
                Y = model(X)
            else:
                model.eval()
                with torch.no_grad():
                    Y = model(X)
            
            loss:torch.Tensor = criterion(Y, T)

            for metric in self.metrics:
                val = metric(Y, T)
                if metric.__name__ not in results:
                    results[metric.__name__] = []
                results[metric.__name__].append(val)
            
            if train:
                loss.backward()
                optimizer.step()

            results['loss'].append(loss.item())
            if return_XYT:
                results['X'].append(X.detach())
                results['Y'].append(Y.detach())
                results['T'].append(T.detach())
        
        for key in results:
            if key == 'X' or key == 'Y' or key == 'T':
                results[key] = torch.cat(results[key], dim=0)
            else:
                results[key] = np.average(results[key])
        
        results['time'] = (datetime.now() - datetime.strptime(self.data['start_time'], '%Y-%m-%d %H:%M:%S.%f')).total_seconds() / 3600
        results['time_per_sample'] = (datetime.now() - start).total_seconds() / len(dataloader.dataset) / dataloader.batch_size

        self.epoch = self.len()

        return results

    def save(self):
        """
        Save data to runs/{id}/data.json
        """
        self.__init_run_dir__()
        with open(f'runs/{self.id}/data.json', 'w') as f:
            json.dump(self.data, f)

    def get_val(self, key:str, phase:str):
        """
        Get last value of key in phase

        Parameters:
        ----------
        key : str
            Key to get
        phase : str
            Phase to get from
        
        Returns:
        -------
        value : float
            Last value of key in phase. If key is not in data, returns np.nan
        """
        if key not in self.data:
            return np.nan
        return self.data[key][phase]['val'][-1]
    
    def get_avg(self, key:str, phase:str):
        """
        Get last average value of key in phase

        Parameters:
        ----------
        key : str
            Key to get
        phase : str
            Phase to get from
        
        Returns:
        -------
        value : float
            Last average value of key in phase. If key is not in data, returns np.nan
        """
        if key not in self.data:
            return np.nan
        return self.data[key][phase]['avg'][-1]
    
    def len(self):
        """
        Get length of longest phase
        """
        l = 0
        for key in self.data:
            if key == 'start_time':
                continue
            for phase in self.data[key]:
                l_temp = len(self.data[key][phase]['val'])
                if l_temp > l:
                    l = l_temp
        return l

    def __init_run_dir__(self):
        self.__directory_runs__ = Path('runs')
        self.__directory_runs__.mkdir(parents=True, exist_ok=True)

        self.directory = self.__directory_runs__.joinpath(self.id)
        self.directory.mkdir(parents=True, exist_ok=True)

        self.directory_plot = self.directory.joinpath('plot')
        self.directory_plot.mkdir(parents=True, exist_ok=True)

        self.file_data = self.directory.joinpath('data.json')

    def __load__(self):
        if self.file_data.exists():
            self.data = json.load(self.file_data.open('r'))
        else:
            self.data = {}
            self.data['start_time'] = datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')
        self.epoch = self.len()

    def __best_state_dict__(self, fname:str, id:str, suffix:str):
        sd_files = glob(f'{self.directory}/*{id}*{suffix}')

        best_acc = 0, None
        for sd_file in sd_files:
            if not '_acc' in sd_file:
                continue
            s_i = sd_file[:-len(suffix)].find('_acc') + 4
            acc = float(sd_file[s_i:-len(suffix)])
            if acc > best_acc[0]:
                best_acc = acc, Path(sd_file).name
        return best_acc

    def save_state_dict(self, state_dict, fname = 'state_dict.pt'):
        """
        Save state_dict to runs/{id}/{fname}

        Parameters:
        ----------
        state_dict : dict
            State dict to save
        fname : str, default = 'state_dict.pt'
            Filename to save to
        """
        self.__init_run_dir__()
        sd = copy.deepcopy(state_dict)
        for k, v in sd.items():
            if hasattr(v, 'cpu'):
                sd[k] = v.cpu()

        file_state_dict = self.directory.joinpath(fname)
        with file_state_dict.open('wb') as f:
            torch.save(sd, f)

    def save_best_state_dict(self, state_dict, new_acc:float, fname = 'state_dict.pt'):
        """
        Save state_dict if new_acc is better than previous best

        Parameters:
        ----------
        state_dict : dict
            State dict to save
        new_acc : float
            New accuracy
        epoch : int, default = None
            Epoch to save
        fname : str, default = 'state_dict.pt'
            Filename to save to
        """
        file_state_dict = self.directory.joinpath(fname)
        suffix = file_state_dict.suffix
        id = file_state_dict.name[:-len(suffix)]
        
        best_acc, best_file = self.__best_state_dict__(fname, id, suffix)

        if new_acc > best_acc:
            fname = f'{id}_e{self.epoch}_acc{new_acc:0.8f}{suffix}'
            self.save_state_dict(state_dict, fname)

    def load_state_dict(self, model:torch.nn.Module, fname = None):
        """
        Load state_dict from runs/{id}/{fname}

        Parameters:
        ----------
        model : torch.nn.Module
            Model to load state_dict into
        fname : str, default = None
            Filename to load from
        """
        if fname is None:
            fname = 'state_dict.pt'
        file_state_dict = self.directory.joinpath(fname)
        if file_state_dict.exists():
            with file_state_dict.open('rb') as f:
                model.load_state_dict(torch.load(f, weights_only = False))
        else:
            try:
                console.warn(f'File runs/{self.id}/{fname} not found.')
            except:
                print(f'File runs/{self.id}/{fname} not found.')
            return
        
    def load_best_state_dict(self, model:torch.nn.Module, fname = 'state_dict.pt', verbose:bool = False):
        """
        Load best state_dict from runs/{id}/{fname}

        Parameters:
        ----------
        model : torch.nn.Module
            Model to load state_dict into
        fname : str, default = 'state_dict.pt'
            Filename to load from
        verbose : bool, default = False
            Print loaded file
        """
        file_state_dict = self.directory.joinpath(fname)
        suffix = file_state_dict.suffix
        id = file_state_dict.name[:-len(suffix)]
        
        best_acc, best_file = self.__best_state_dict__(fname, id, suffix)
        if best_file is not None:
            self.load_state_dict(model, best_file)
            if not verbose:
                try:
                    console.success(f'Loaded {best_file}')
                except:
                    print(f'Loaded {best_file}')

    def pickle_dump(self, model:torch.nn.Module, fname = 'model.pkl'):
        """
        Pickle model to runs/{id}/{fname}

        Parameters:
        ----------
        model : torch.nn.Module
            Model to pickle
        fname : str, default = 'model.pkl'
            Filename to save to
        """
        self.__init_run_dir__()

        file_pkl = self.directory.joinpath(fname)
        with file_pkl.open('wb') as f:
            pkl.dump(model, f)

    def pickle_load(self, fname = 'model.pkl'):
        """
        Load model from runs/{id}/{fname}

        Parameters:
        ----------
        fname : str, default = 'model.pkl'
            Filename to load from
        """
        file_pkl = self.directory.joinpath(fname)
        if file_pkl.exists():
            with open(file_pkl, 'rb') as f:
                model = pkl.load(f)
        else:
            try:
                console.warn(f'File runs/{self.id}/{fname} not found.')
            except:
                print(f'File runs/{self.id}/{fname} not found.')
            return
        return model
    
if __name__ == '__main__':
    t1 = datetime.strftime(datetime.now(), '%Y-%m-%d %H:%M:%S.%f')
    test = datetime.strptime('2025-02-28 14:31:23.3234', '%Y-%m-%d %H:%M:%S.%f')

    run = Run('test')
    run.save_best_state_dict(torch.nn.Linear(10, 10).state_dict(), 0.919342, 80)
    run.save_best_state_dict(torch.nn.Linear(10, 10).state_dict(), 0.899342, 81)
    run.save_best_state_dict(torch.nn.Linear(10, 10).state_dict(), 0.999342, 82)

    run.load_best_state_dict(torch.nn.Linear(10, 10))
    pass


import urllib.request
from torch.utils.data import Dataset, DataLoader
from torch.utils.data.sampler import Sampler
from pathlib import Path
from platformdirs import user_cache_dir
from tqdm import tqdm
from glob import glob
from threading import Thread
from huggingface_hub import hf_hub_download, list_repo_files
import xml.etree.ElementTree as ET
import numpy as np
import os
import json
import pkg_resources
import urllib
import tarfile
import cv2 as cv
import csv
import torch
import rsp.ml.multi_transforms.multi_transforms as multi_transforms
import time
import pandas as pd
import gdown
import shutil
import zipfile

try:
    import rsp.common.console as console
except Exception as e:
    print(e)

#__example__ from rsp.ml.dataset import TUCRID
#__example__ from rsp.ml.dataset import ReplaceBackgroundRGBD
#__example__ import rsp.ml.multi_transforms as multi_transforms
#__example__ import cv2 as cv
#__example__
#__example__ backgrounds = TUCRID.load_backgrounds_color()
#__example__ transforms = multi_transforms.Compose([
#__example__     ReplaceBackgroundRGBD(backgrounds),
#__example__     multi_transforms.Stack()
#__example__ ])
#__example__ 
#__example__ ds = TUCRID('train', transforms=transforms)
#__example__ 
#__example__ for X, T in ds:
#__example__   for x in X.permute(0, 2, 3, 1):
#__example__     img_color = x[:, :, :3].numpy()
#__example__     img_depth = x[:, :, 3].numpy()
#__example__ 
#__example__     cv.imshow('color', img_color)
#__example__     cv.imshow('depth', img_depth)
#__example__ 
#__example__     cv.waitKey(30)
class TUCRID(Dataset):
    """
    Dataset class for the Robot Interaction Dataset by University of Technology Chemnitz (TUCRID).
    """
    REPO_ID = 'SchulzR97/TUCRID'
    CACHE_DIRECTORY = Path(user_cache_dir('rsp-ml', 'Robert Schulz')).joinpath('dataset', 'TUCRID')
    COLOR_DIRECTORY = CACHE_DIRECTORY.joinpath('color')
    DEPTH_DIRECTORY = CACHE_DIRECTORY.joinpath('depth')
    BACKGROUND_DIRECTORY = CACHE_DIRECTORY.joinpath('background')
    PHASES = ['train', 'val']

    def __init__(
            self,
            phase:str,
            load_depth_data:bool = True,
            sequence_length:int = 30,
            num_classes:int = 10,
            transforms:multi_transforms.Compose = multi_transforms.Compose([]),
            cache_dir:str = None
    ):
        """
        Initializes a new instance.

        Parameters
        ----------
        phase : str
            Dataset phase [train|val]
        load_depth_data : bool, default = True
            Load depth data
        sequence_length : int, default = 30
            Length of the sequences
        num_classes : int, default = 10
            Number of classes
        transforms : rsp.ml.multi_transforms.Compose = default = rsp.ml.multi_transforms.Compose([])
            Transformations, that will be applied to each input sequence. See documentation of `rsp.ml.multi_transforms` for more details.
        """
        assert phase in TUCRID.PHASES, f'Phase "{phase}" not in {TUCRID.PHASES}'

        if cache_dir is not None:
            TUCRID.CACHE_DIRECTORY = Path(cache_dir)
            TUCRID.COLOR_DIRECTORY = TUCRID.CACHE_DIRECTORY.joinpath('color')
            TUCRID.DEPTH_DIRECTORY = TUCRID.CACHE_DIRECTORY.joinpath('depth')
            TUCRID.BACKGROUND_DIRECTORY = TUCRID.CACHE_DIRECTORY.joinpath('background')

        self.phase = phase
        self.load_depth_data = load_depth_data
        self.sequence_length = sequence_length
        self.num_classes = num_classes
        self.transforms = transforms

        self.__download__()
        self.__load__()
        pass

    def __len__(self):
        return len(self.sequences)
    
    def __getitem__(self, idx):
        id = self.sequences['id'][idx]
        action = self.sequences['action'][idx]
        link = self.sequences['link'][idx]

        color_files = sorted(glob(f'{TUCRID.COLOR_DIRECTORY}/{link}/*.jpg'))
        assert len(color_files) >= self.sequence_length, f'Not enough frames for {id}.'

        if len(color_files) > self.sequence_length:
            start_idx = np.random.randint(0, len(color_files) - self.sequence_length)
            end_idx = start_idx + self.sequence_length
        else:
            start_idx = 0
            end_idx = start_idx + self.sequence_length

        color_images = []
        depth_images = []
        for color_file in color_files[start_idx:end_idx]:

            color_file = Path(color_file)

            img = cv.imread(str(color_file))
            color_images.append(img)

            if self.load_depth_data:
                depth_file = TUCRID.DEPTH_DIRECTORY.joinpath(f'{link}/{color_file.name}')
                img = cv.imread(str(depth_file), cv.IMREAD_UNCHANGED)
                depth_images.append(img)
        
        X = torch.tensor(np.array(color_images), dtype=torch.float32) / 255
        if self.load_depth_data:
            X_depth = torch.tensor(np.array(depth_images), dtype=torch.float32).unsqueeze(3) / 255
            X = torch.cat([X, X_depth], dim=3)
        X = X.permute(0, 3, 1, 2)
        T = torch.zeros((self.num_classes), dtype=torch.float32)
        T[action] = 1

        self.transforms.__reset__()
        X = self.transforms(X)
        
        return X, T

    def __download__(self):                        
        TUCRID.CACHE_DIRECTORY.mkdir(exist_ok=True, parents=True)

        TUCRID.__download_metadata__()

        TUCRID.__download_backgrounds__()

        TUCRID.__download_sequences__(self.load_depth_data)

    def __download_file__(filename, retries = 10):
        attempts = 0
        while True:
            try:
                hf_hub_download(
                    repo_id=TUCRID.REPO_ID,
                    repo_type='dataset',
                    local_dir=TUCRID.CACHE_DIRECTORY,
                    filename=str(filename)
                )
                break
            except Exception as e:
                if attempts < retries:
                    attempts += 1
                else:
                    raise e

    def __download_metadata__():
        for phase in TUCRID.PHASES:
            if not f'{phase}.json' in os.listdir(TUCRID.CACHE_DIRECTORY):
                TUCRID.__download_file__(f'{phase}.json')

    def __download_backgrounds__():
        # color
        background_color_dir = TUCRID.BACKGROUND_DIRECTORY.joinpath('color')
        if not background_color_dir.exists() or len(os.listdir(background_color_dir)) == 0:
            TUCRID.__download_file__('background/color.tar.gz')
            background_color_tarfile = TUCRID.BACKGROUND_DIRECTORY.joinpath('color.tar.gz')
            with tarfile.open(background_color_tarfile, 'r:gz') as tar:
                tar.extractall(background_color_dir)
            os.remove(background_color_tarfile)

        # depth
        background_depth_dir = TUCRID.BACKGROUND_DIRECTORY.joinpath('depth')
        if not background_depth_dir.exists() or len(os.listdir(background_depth_dir)) == 0:
            TUCRID.__download_file__('background/depth.tar.gz')
            background_depth_tarfile = TUCRID.BACKGROUND_DIRECTORY.joinpath('depth.tar.gz')
            with tarfile.open(background_depth_tarfile, 'r:gz') as tar:
                tar.extractall(background_depth_dir)
            os.remove(background_depth_tarfile)

    def __download_sequences__(load_depth_data):
        repo_files = [Path(file) for file in list_repo_files(TUCRID.REPO_ID, repo_type='dataset')]
        color_files = [file for file in repo_files if file.parent.name == 'color']

        prog = tqdm(color_files, leave=False)
        for color_file in prog:
            prog.set_description(f'Downloading {color_file}')
            local_dir = TUCRID.COLOR_DIRECTORY.joinpath(color_file.name.replace('.tar.gz', ''))
            if local_dir.exists() and len(os.listdir(local_dir)) > 0:
                continue
            TUCRID.__download_file__(color_file)
            tar_color = TUCRID.COLOR_DIRECTORY.joinpath(color_file.name)
            with tarfile.open(tar_color, 'r:gz') as tar:
                tar.extractall(local_dir)
            os.remove(tar_color)

        if load_depth_data:
            depth_files = [file for file in repo_files if file.parent.name == 'depth']
            prog = tqdm(depth_files)
            for depth_file in prog:
                prog.set_description(f'Downloading {depth_file}')
                local_dir = TUCRID.DEPTH_DIRECTORY.joinpath(depth_file.name.replace('.tar.gz', ''))
                if local_dir.exists() and len(os.listdir(local_dir)) > 0:
                    continue
                TUCRID.__download_file__(depth_file)
                tar_depth = TUCRID.DEPTH_DIRECTORY.joinpath(depth_file.name)
                with tarfile.open(tar_depth, 'r:gz') as tar:
                    tar.extractall(local_dir)
                os.remove(tar_depth)

    def __load__(self):
        with open(TUCRID.CACHE_DIRECTORY.joinpath(f'{self.phase}.json'), 'r') as f:
            self.sequences = pd.DataFrame(json.load(f))

        self.action_labels = self.sequences[['action', 'label']].drop_duplicates().sort_values('action')['label'].tolist()

    def get_uniform_sampler(self):
        groups = self.sequences.groupby('action')

        action_counts = groups.size().to_numpy()
        action_weights = 1. / (action_counts / action_counts.sum())
        action_weights = action_weights / action_weights.sum()

        for action, prob in enumerate(action_weights):
            self.sequences.loc[self.sequences['action'] == action, 'sample_prob'] = prob

        class UniformSampler(Sampler):
            def __init__(self, probs, len):
                self.probs = probs
                self.len = len

            def __iter__(self):
                indices = torch.tensor(self.probs).multinomial(self.len, replacement=True).tolist()
                for idx in indices:
                    yield idx

            def __len__(self):
                return self.len
            
        return UniformSampler(self.sequences['sample_prob'].to_numpy(), len(self.sequences))

    def load_backgrounds(load_depth_data:bool = True):
        """
        Loads the background images.

        Parameters
        ----------
        load_depth_data : bool, default = True
            If set to `True`, the depth images will be loaded as well.
        """
        bg_color_dir = TUCRID.BACKGROUND_DIRECTORY.joinpath('color')
        bg_depth_dir = TUCRID.BACKGROUND_DIRECTORY.joinpath('depth')

        if not bg_color_dir.exists() or len(os.listdir(bg_color_dir)) == 0:
            TUCRID.__download_backgrounds__()
        if load_depth_data and (not bg_depth_dir.exists() or len(os.listdir(bg_depth_dir)) == 0):
            TUCRID.__download_backgrounds__()

        bg_color_files = sorted(glob(f'{bg_color_dir}/*'))

        backgrounds = []
        for fname_color in bg_color_files:
            fname_color = Path(fname_color)
            bg_color = cv.imread(str(fname_color))

            if load_depth_data:
                fname_depth = TUCRID.BACKGROUND_DIRECTORY.joinpath('depth', fname_color.name.replace('_color', '_depth'))
                bg_depth = cv.imread(str(fname_depth), cv.IMREAD_UNCHANGED)
                backgrounds.append((bg_color, bg_depth))
            else:
                backgrounds.append((bg_color,))
        return backgrounds

class TUCHRI(Dataset):
    """
    Dataset class for the Robot Interaction Dataset by University of Technology Chemnitz (TUCHRI).
    """
    REPO_ID = 'SchulzR97/TUC-HRI'
    CACHE_DIRECTORY = Path(user_cache_dir('rsp-ml', 'Robert Schulz')).joinpath('dataset', 'TUCHRI')
    COLOR_DIRECTORY = CACHE_DIRECTORY.joinpath('color')
    DEPTH_DIRECTORY = CACHE_DIRECTORY.joinpath('depth')
    PHASES = ['train', 'val']

    def __init__(
            self,
            phase:str,
            load_depth_data:bool = True,
            sequence_length:int = 30,
            num_classes:int = 10,
            transforms:multi_transforms.Compose = multi_transforms.Compose([]),
            cache_dir:str = None
    ):
        """
        Initializes a new instance.

        Parameters
        ----------
        phase : str
            Dataset phase [train|val]
        load_depth_data : bool, default = True
            Load depth data
        sequence_length : int, default = 30
            Length of the sequences
        num_classes : int, default = 10
            Number of classes
        transforms : rsp.ml.multi_transforms.Compose = default = rsp.ml.multi_transforms.Compose([])
            Transformations, that will be applied to each input sequence. See documentation of `rsp.ml.multi_transforms` for more details.
        """
        assert phase in TUCRID.PHASES, f'Phase "{phase}" not in {TUCRID.PHASES}'

        if cache_dir is not None:
            TUCHRI.CACHE_DIRECTORY = Path(cache_dir)
            TUCHRI.COLOR_DIRECTORY = TUCHRI.CACHE_DIRECTORY.joinpath('color')
            TUCHRI.DEPTH_DIRECTORY = TUCHRI.CACHE_DIRECTORY.joinpath('depth')

        self.phase = phase
        self.load_depth_data = load_depth_data
        self.sequence_length = sequence_length
        self.num_classes = num_classes
        self.transforms = transforms

        #self.__download__()
        self.__load__()
        pass

    def __len__(self):
        return len(self.sequences)
    
    def __getitem__(self, idx):
        id = self.sequences['id'][idx]
        action = self.sequences['action'][idx]
        link = self.sequences['link'][idx]

        color_files = sorted(glob(f'{TUCHRI.COLOR_DIRECTORY}/{link}/*.jpg'))
        assert len(color_files) >= self.sequence_length, f'Not enough frames for {id}.'

        if len(color_files) > self.sequence_length:
            start_idx = np.random.randint(0, len(color_files) - self.sequence_length)
            end_idx = start_idx + self.sequence_length
        else:
            start_idx = 0
            end_idx = start_idx + self.sequence_length

        color_images = []
        depth_images = []
        for color_file in color_files[start_idx:end_idx]:

            color_file = Path(color_file)

            img = cv.imread(str(color_file))
            color_images.append(img)

            if self.load_depth_data:
                depth_file = TUCHRI.DEPTH_DIRECTORY.joinpath(f'{link}/{color_file.name}')
                img = cv.imread(str(depth_file), cv.IMREAD_UNCHANGED)
                depth_images.append(img)
        
        X = torch.tensor(np.array(color_images), dtype=torch.float32) / 255
        if self.load_depth_data:
            X_depth = torch.tensor(np.array(depth_images), dtype=torch.float32).unsqueeze(3) / 255
            X = torch.cat([X, X_depth], dim=3)
        X = X.permute(0, 3, 1, 2)
        T = torch.zeros((self.num_classes), dtype=torch.float32)
        T[action] = 1

        self.transforms.__reset__()
        X = self.transforms(X)
        
        return X, T

    def __download__(self):                        
        TUCHRI.CACHE_DIRECTORY.mkdir(exist_ok=True, parents=True)

        TUCHRI.__download_metadata__()

        TUCHRI.__download_sequences__(self.load_depth_data)

    def __download_file__(filename, retries = 10):
        attempts = 0
        while True:
            try:
                hf_hub_download(
                    repo_id=TUCHRI.REPO_ID,
                    repo_type='dataset',
                    local_dir=TUCHRI.CACHE_DIRECTORY,
                    filename=str(filename)
                )
                break
            except Exception as e:
                if attempts < retries:
                    attempts += 1
                else:
                    raise e

    def __download_metadata__():
        for phase in TUCHRI.PHASES:
            if not f'{phase}.json' in os.listdir(TUCHRI.CACHE_DIRECTORY):
                TUCHRI.__download_file__(f'{phase}.json')

    def __download_sequences__(load_depth_data):
        repo_files = [Path(file) for file in list_repo_files(TUCHRI.REPO_ID, repo_type='dataset')]
        color_files = [file for file in repo_files if file.parent.name == 'color']

        prog = tqdm(color_files, leave=False)
        for color_file in prog:
            prog.set_description(f'Downloading {color_file}')
            local_dir = TUCHRI.COLOR_DIRECTORY.joinpath(color_file.name.replace('.tar.gz', ''))
            if local_dir.exists() and len(os.listdir(local_dir)) > 0:
                continue
            TUCHRI.__download_file__(color_file)
            tar_color = TUCHRI.COLOR_DIRECTORY.joinpath(color_file.name)
            with tarfile.open(tar_color, 'r:gz') as tar:
                tar.extractall(local_dir)
            os.remove(tar_color)

        if load_depth_data:
            depth_files = [file for file in repo_files if file.parent.name == 'depth']
            prog = tqdm(depth_files)
            for depth_file in prog:
                prog.set_description(f'Downloading {depth_file}')
                local_dir = TUCHRI.DEPTH_DIRECTORY.joinpath(depth_file.name.replace('.tar.gz', ''))
                if local_dir.exists() and len(os.listdir(local_dir)) > 0:
                    continue
                TUCHRI.__download_file__(depth_file)
                tar_depth = TUCHRI.DEPTH_DIRECTORY.joinpath(depth_file.name)
                with tarfile.open(tar_depth, 'r:gz') as tar:
                    tar.extractall(local_dir)
                os.remove(tar_depth)

    def __load__(self):
        with open(TUCHRI.CACHE_DIRECTORY.joinpath(f'{self.phase}.json'), 'r') as f:
            self.sequences = pd.DataFrame(json.load(f))

        self.action_labels = self.sequences[['action', 'label']].drop_duplicates().sort_values('action')['label'].tolist()

    def get_uniform_sampler(self):
        groups = self.sequences.groupby('action')

        action_counts = groups.size().to_numpy()
        action_weights = 1. / (action_counts / action_counts.sum())
        action_weights = action_weights / action_weights.sum()

        for action, prob in enumerate(action_weights):
            self.sequences.loc[self.sequences['action'] == action, 'sample_prob'] = prob

        class UniformSampler(Sampler):
            def __init__(self, probs, len):
                self.probs = probs
                self.len = len

            def __iter__(self):
                indices = torch.tensor(self.probs).multinomial(self.len, replacement=True).tolist()
                for idx in indices:
                    yield idx

            def __len__(self):
                return self.len
            
        return UniformSampler(self.sequences['sample_prob'].to_numpy(), len(self.sequences))

    def load_backgrounds(load_depth_data:bool = True):
        """
        Loads the background images.

        Parameters
        ----------
        load_depth_data : bool, default = True
            If set to `True`, the depth images will be loaded as well.
        """
        bg_color_dir = TUCHRI.BACKGROUND_DIRECTORY.joinpath('color')
        bg_depth_dir = TUCHRI.BACKGROUND_DIRECTORY.joinpath('depth')

        if not bg_color_dir.exists() or len(os.listdir(bg_color_dir)) == 0:
            TUCHRI.__download_backgrounds__()
        if load_depth_data and (not bg_depth_dir.exists() or len(os.listdir(bg_depth_dir)) == 0):
            TUCHRI.__download_backgrounds__()

        bg_color_files = sorted(glob(f'{bg_color_dir}/*'))

        backgrounds = []
        for fname_color in bg_color_files:
            fname_color = Path(fname_color)
            bg_color = cv.imread(str(fname_color))

            if load_depth_data:
                fname_depth = TUCHRI.BACKGROUND_DIRECTORY.joinpath('depth', fname_color.name.replace('_color', '_depth'))
                bg_depth = cv.imread(str(fname_depth), cv.IMREAD_UNCHANGED)
                backgrounds.append((bg_color, bg_depth))
            else:
                backgrounds.append((bg_color,))
        return backgrounds

#__example__ from rsp.ml.dataset import HMDB51
#__example__ import rsp.ml.multi_transforms as multi_transforms
#__example__ import cv2 as cv
#__example__ 
#__example__ transforms = multi_transforms.Compose([
#__example__     multi_transforms.Color(1.5, p=0.5),
#__example__     multi_transforms.Stack()
#__example__ ])
#__example__ ds = HMDB51('train', fold=1, transforms=transforms)
#__example__ 
#__example__ for X, T in ds:
#__example__   for x in X.permute(0, 2, 3, 1):
#__example__     img_color = x[:, :, :3].numpy()
#__example__     img_depth = x[:, :, 3].numpy()
#__example__ 
#__example__     cv.imshow('color', img_color)
#__example__     cv.imshow('depth', img_depth)
#__example__ 
#__example__     cv.waitKey(30)
class HMDB51(Dataset):
    """
    Dataset class for HMDB51.
    """
    def __init__(
            self,
            split:str,
            fold:int = None,
            cache_dir:str = None,
            force_reload:bool = False,
            target_size = (400, 400),
            sequence_length:int = 30,
            transforms:multi_transforms.Compose = multi_transforms.Compose([]),
            verbose:bool = True
    ):
        """
        Initializes a new instance.

        Parameters
        ----------
        split : str
            Dataset split [train|val|test]
        fold : int
            Fold number. The dataset is split into 3 folds. If fold is None, all folds will be loaded.
        cache_dir : str, default = None
            Directory to store the downloaded files. If set to `None`, the default cache directory will be used
        force_reload : bool, default = False
            If set to `True`, the dataset will be reloaded
        target_size : (int, int), default = (400, 400)
            Size of the frames. The frames will be resized to this size.
        sequence_length : int, default = 30
            Length of the sequences
        transforms : rsp.ml.multi_transforms.Compose = default = rsp.ml.multi_transforms.Compose([])
            Transformations, that will be applied to each input sequence. See documentation of `rsp.ml.multi_transforms` for more details.
        verbose : bool, default = False
            If set to `True`, the progress will be printed.
        """
        self.download_link = f'https://drive.google.com/uc?id=1iMQo02o9iEuawhGcicBvzqbZxvtoLCok'
        self.split = split
        self.fold = fold
        self.force_reload = force_reload
        self.target_size = target_size
        self.sequence_length = sequence_length
        self.transforms = transforms
        self.verbose = verbose

        if cache_dir is None:
            self.__cache_dir__ = Path(user_cache_dir("rsp-ml", "Robert Schulz")).joinpath('dataset', 'HMDB51')
        else:
            self.__cache_dir__ = Path(cache_dir).joinpath('HMDB51')
        self.__cache_dir__.mkdir(parents=True, exist_ok=True)

        self.__download__()
        self.__list_files__()

    def __len__(self):
        return len(self.__files__)
    
    def __getitem__(self, index):
        action, fname = self.__files__[index]

        cap = cv.VideoCapture(fname)
        cnt = int(cap.get(cv.CAP_PROP_FRAME_COUNT))
        width = cap.get(cv.CAP_PROP_FRAME_WIDTH)
        height = cap.get(cv.CAP_PROP_FRAME_HEIGHT)

        if cnt-self.sequence_length <= 0:
            start_idx = 0
        else:
            start_idx = np.random.randint(0, cnt-self.sequence_length)
        end_idx = start_idx + self.sequence_length

        frames = []
        cap.set(cv.CAP_PROP_POS_FRAMES, start_idx)
        while True:
            ret, frame = cap.read()
            if not ret:
                break
            frame = cv.resize(frame, self.target_size)
            frames.append(frame)
            if len(frames) >= self.sequence_length:
                break

        X = torch.tensor(np.array(frames), dtype=torch.float32).permute(0, 3, 1, 2) / 255
        T = torch.zeros((len(self.action_labels)), dtype=torch.float32)
        T[action] = 1

        if X.shape[0] < self.sequence_length:
            if self.verbose:
                try:
                    console.warn(f'Seuqnce length was {X.shape[0]}. Expected {self.sequence_length}. Automatic expanding...')
                except:
                    print(f'Seuqnce length was {X.shape[0]}. Expected {self.sequence_length}. Automatic expanding...')
            X = torch.concat([X, torch.zeros((self.sequence_length-X.shape[0], X.shape[1], X.shape[2], X.shape[3]), dtype=torch.float32)])

        X = self.transforms(X)

        return X, T

    def __download__(self):
        zip_file = f'{self.__cache_dir__.parent}/HMDB51.zip'
        if not os.path.isdir(f'{self.__cache_dir__}') or len(os.listdir(self.__cache_dir__)) == 0 or self.force_reload:
            if not os.path.isfile(zip_file):
                if self.verbose:
                    try:
                        console.print_c('Downloading HMDB51 dataset...', color=console.color.GREEN)
                    except:
                        print('Downloading HMDB51 dataset...')
                gdown.download(self.download_link, zip_file, quiet=not self.verbose)
            with zipfile.ZipFile(zip_file, 'r') as zip_ref:
                try:
                    console.print_c('Extracting zip file...', color=console.color.GREEN)
                except:
                    print('Extracting zip file...')
                zip_ref.extractall(self.__cache_dir__.parent)
            if os.path.isdir(f'{self.__cache_dir__.parent}/__MACOSX'):
                shutil.rmtree(f'{self.__cache_dir__.parent}/__MACOSX')
            os.remove(zip_file)

    def __list_files__(self):
        self.__files__ = []

        self.action_labels = sorted([Path(folder).name for folder in glob(f'{self.__cache_dir__}/sequences/*')])

        if self.fold is None:
            split_files = glob(f'{self.__cache_dir__}/splits/*_split*.txt')
        else:
            split_files = glob(f'{self.__cache_dir__}/splits/*_split{self.fold}.txt')

        for split_file in split_files:
            split_file = Path(split_file)
            end_idx = split_file.name.find('_test_split')
            label = split_file.name[:end_idx]
            action_idx = self.action_labels.index(label)

            with open(split_file, 'r') as file:
                lines = [line for line in file.read().split('\n') if len(line) > 0]
                for line in lines:
                    fname = line.split(' ')[0]
                    fname = f'{self.__cache_dir__}/sequences/{label}/{fname}'
                    split_idx = int(line.split(' ')[1])
                    if self.split == 'train' and split_idx == 1:
                        self.__files__.append((action_idx, fname))
                    elif self.split == 'val' and split_idx == 2:
                        self.__files__.append((action_idx, fname))
                    elif self.split == 'test' and split_idx == 0:
                        self.__files__.append((action_idx, fname))
                pass
            pass

#__example__ from rsp.ml.dataset import Kinetics
#__example__ 
#__example__ ds = Kinetics(split='train', type=400)
#__example__
#__example__ for X, T in ds:
#__example__     print(X)
class Kinetics(Dataset):
    """
    Dataset class for the Kinetics dataset.
    """
    def __init__(
        self,
        split:str,
        sequence_length:int = 60,
        type:int = 400,
        frame_size = (400, 400),
        transforms:multi_transforms.Compose = multi_transforms.Compose([]),
        cache_dir:str = None,
        num_threads:int = 0,
        verbose:bool = True
    ):
        """
        Initializes a new instance.
        
        Parameters
        ----------
        split : str
            Dataset split [train|val]
        sequence_length : int, default = 60
            Length of the sequences
        type : int, default = 400
            Type of the kineticts dataset. Currently only 400 is supported.
        frame_size : (int, int), default = (400, 400)
            Size of the frames. The frames will be resized to this size.
        transforms : rsp.ml.multi_transforms.Compose = default = rsp.ml.multi_transforms.Compose([])
            Transformations, that will be applied to each input sequence. See documentation of `rsp.ml.multi_transforms` for more details.
        cache_dir : str, default = None
            Directory to store the downloaded files. If set to `None`, the default cache directory will be used
        num_threads : int, default = 0
            Number of threads to use for downloading the files.
        verbose : bool, default = True
            If set to `True`, the progress and additional information will be printed.
        """
        super().__init__()

        assert split in ['train', 'val'], f'{split} is not a valid split.'
        assert type in [400], f'{type} is not a valid type.'

        self.split = split
        self.type = type
        self.frame_size = frame_size
        self.sequence_length = sequence_length
        self.transforms = transforms
        self.num_threads = num_threads
        self.verbose = verbose

        if cache_dir is None:
            self.__cache_dir__ = Path(user_cache_dir("rsp-ml", "Robert Schulz")).joinpath('dataset', f'KINETICS{type}')
        else:
            self.__cache_dir__ = Path(cache_dir)
        self.__cache_dir__.mkdir(parents=True, exist_ok=True)

        self.__toTensor__ = multi_transforms.ToTensor()
        self.__stack__ = multi_transforms.Stack()

        self.__download__()
        self.__annotations__, self.action_labels = self.__load_annotations_labels__()
        self.__invalid_files__ = self.__get_invalid_files__()
        self.__files__ = self.__list_files__()

    def __getitem__(self, index):
        youtube_id, fname = self.__files__[index]

        annotation = self.__annotations__[youtube_id]

        if annotation['time_end'] - annotation['time_start'] > self.sequence_length:
            start_idx = np.random.randint(annotation['time_start'], annotation['time_end']-self.sequence_length)
            end_idx = start_idx + self.sequence_length
        else:
            start_idx = (annotation['time_end'] - annotation['time_start'])//2 - self.sequence_length//2
            end_idx = (annotation['time_end'] - annotation['time_start'])//2 + self.sequence_length//2

        cap = cv.VideoCapture(fname)
        cap.set(cv.CAP_PROP_POS_FRAMES, start_idx)

        frames = []
        while True:
            ret, frame = cap.read()
            if not ret:
                break
            frame = cv.resize(frame, self.frame_size)
            frames.append(frame)
            if len(frames) >= end_idx - start_idx:
                break
        frames = np.array(frames) / 255

        if len(frames) == 0:
            X = torch.zeros((self.sequence_length, 3, *self.frame_size), dtype=torch.float32)
            if self.verbose:
                console.warn(f'No frames found for {youtube_id}.')
        else:
            X = torch.tensor(frames, dtype=torch.float32).permute(0, 3, 1, 2)
        T = torch.zeros((len(self.action_labels)), dtype=torch.float32)
        cls = self.action_labels.index(annotation['label'])
        T[cls] = 1

        if X.shape[0] < self.sequence_length:
            if self.verbose:
                console.warn(f'Seuqnce length was {X.shape[0]}. Expected {self.sequence_length}. Automatic expanding...')
            X = torch.concat([X, torch.zeros((self.sequence_length-X.shape[0], X.shape[1], X.shape[2], X.shape[3]), dtype=torch.float32)])

        X = self.transforms(X)

        return X, T
    
    def __get_invalid_files__(self):
        valid_files_file = self.__cache_dir__.joinpath('valid_files.txt')
        invalid_files_file = self.__cache_dir__.joinpath('invalid_files.txt')

        if valid_files_file.exists():
            with open(str(valid_files_file), 'r') as file:
                lines = file.read().split('\n')
            valid_files = [line for line in lines if len(line) > 0]
        else:
            valid_files = []  

        if invalid_files_file.exists():
            with open(str(invalid_files_file), 'r') as file:
                lines = file.read().split('\n')
            invalid_files = [line for line in lines if len(line) > 0]
            # invalid_files = list(dict.fromkeys(invalid_files))
            # with open(str(invalid_files_file), 'w') as file:
            #     for line in invalid_files:
            #         file.write(f'{line}\n')
        else:
            invalid_files = []    

        videos_dir = self.__cache_dir__.joinpath('videos', self.split)
        links = glob(f'{videos_dir}/k{self.type}*/*.mp4')

        prog = tqdm(links, leave=False)
        for i, link in enumerate(prog):
            file = Path(link)
            link = f'/{file.parent.parent.parent.name}/{file.parent.parent.name}/{file.parent.name}/{file.name}'
            youtube_id = file.name[:-18]

            if youtube_id in invalid_files or youtube_id in valid_files:
                continue

            if not file.exists():
                with open(invalid_files_file, 'a') as file:
                    file.write(f'{youtube_id}\n')
                invalid_files.append(youtube_id)
                continue

            cap = cv.VideoCapture(str(file))
            if not cap.isOpened():
                with open(invalid_files_file, 'a') as file:
                    file.write(f'{youtube_id}\n')
                invalid_files.append(youtube_id)
                cap.release()
                continue

            idx_start = int(file.name[-17:-11])
            idx_end = int(file.name[-10:-4])
            cnt = int(cap.get(cv.CAP_PROP_FRAME_COUNT))

            ret, _ = cap.read()

            if not ret:
                with open(invalid_files_file, 'a') as file:
                    file.write(f'{youtube_id}\n')
                invalid_files.append(youtube_id)
                cap.release()
                continue

            cap.release()

            with open(valid_files_file, 'a') as file:
                file.write(f'{youtube_id}\n')
            valid_files.append(youtube_id)

            prog.set_description(f'Invalid files {len(invalid_files)} ({len(invalid_files)/(len(invalid_files)+len(valid_files))*100:.2f}%)')
            #break

        return invalid_files

    def __len__(self):
        return len(self.__files__)

    def __get_labels__(self):
        labels = {}
        df = pd.DataFrame(self.__annotations__)
        for i, (key, _) in enumerate(df.groupby('label')):
            key = key.replace('"', '')
            labels[key] = i
        return labels

    def __download__(self):
        def get_fname_resource(resource_name):
            fname = pkg_resources.resource_filename('rsp', resource_name)
            return Path(fname)
        
        def download_file(link, fname, retries = 10):
            attempt = 0
            while attempt < retries:
                try:
                    urllib.request.urlretrieve(link, fname)
                    break
                except urllib.error.ContentTooShortError as e:
                    attempt += 1
                except Exception as e:
                    attempt += 1

        def unpack(src, dest, remove = True):
            with tarfile.open(src, "r:gz") as tar:
                tar.extractall(path=dest)
            if remove:
                os.remove(src)

        anno_link_file = get_fname_resource(f'ml/dataset/links/kinetics/annotations/k{self.type}_annotations.txt')
        with open(anno_link_file, 'r') as file:
            links = file.read().split('\n')
            cache_anno_dir = Path(self.__cache_dir__).joinpath('annotations')
            cache_anno_dir.mkdir(parents=True, exist_ok=True)
            for link in links:
                fname = link.split('/')[-1]
                fname = cache_anno_dir.joinpath(f'k{self.type}_{fname}')
                if fname.exists():
                    continue
                download_file(link, fname)

        path_link_files = [
            get_fname_resource(f'ml/dataset/links/kinetics/paths/k{self.type}_train_path.txt'),
            get_fname_resource(f'ml/dataset/links/kinetics/paths/k{self.type}_test_path.txt'),
            get_fname_resource(f'ml/dataset/links/kinetics/paths/k{self.type}_val_path.txt')
        ]

        cache_archives_dir = self.__cache_dir__.joinpath('archives')
        cache_archives_dir.mkdir(parents=True, exist_ok=True)

        cache_videos_dir = self.__cache_dir__.joinpath('videos')
        cache_videos_dir.mkdir(parents=True, exist_ok=True)

        threads = []

        prog1 = tqdm(path_link_files, leave=False)
        for link_file in prog1:
            prog1.set_description(f'Downloading {link_file.stem}')

            with open(link_file, 'r') as file:
                links = file.read().split('\n')
            prog2 = tqdm(links, leave=False)
            for link in prog2:
                prog2.set_description(link)

                def process_link(link):
                    split, fname = link.split('/')[-2:]

                    video_dir = cache_videos_dir.joinpath(split, 'k' + str(self.type) + '_' + fname.split(".")[0])
                    if video_dir.exists():
                        #continue
                        return

                    archive_file = cache_archives_dir.joinpath(split, f'k{self.type}_{fname}')
                    archive_file.parent.mkdir(parents=True, exist_ok=True)
                    if not archive_file.exists():
                        download_file(link, archive_file)

                    video_dir.mkdir(parents=True, exist_ok=True)
                    try:
                        unpack(archive_file, video_dir, remove=True)
                    except Exception as e:
                        video_dir.rmdir()
                        os.remove(archive_file)
                        download_file(link, archive_file)
                        unpack(archive_file, video_dir, remove=True)

                if self.num_threads == 0:
                    process_link(link)
                else:
                    thread = Thread(target=process_link, args=(link,))
                    while len(threads) >= self.num_threads:
                        threads = [t for t in threads if t.is_alive()]
                        time.sleep(0.1)
                    thread.start()
                    threads.append(thread)

    def __load_annotations_labels__(self):
        annotations_file = self.__cache_dir__.joinpath('annotations', f'k{self.type}_{self.split}.csv')
        annotations = {}
        labels = []
        with open(annotations_file, newline='') as csvfile:
            spamreader = csv.reader(csvfile, delimiter=',', quotechar='|')
            for i, row in enumerate(spamreader):
                if i == 0:
                    continue
                label, youtube_id, time_start, time_end, split, is_cc = row[0], row[1], int(row[2]), int(row[3]), row[4], int(row[5])
                label = label.replace('"', '')
                annotations[youtube_id] = {
                    'label': label,
                    #'youtube_id': youtube_id,
                    'time_start': time_start,
                    'time_end': time_end,
                    'split': split,
                    'is_cc': is_cc
                }
                if label not in labels:
                    labels.append(label)
        return annotations, sorted(labels)

    def __list_files__(self):
        videos_dir = self.__cache_dir__.joinpath('videos', self.split)
        links = glob(f'{videos_dir}/k{self.type}*/*.mp4')
        files = []#{}
        for link in links:
            youtube_id = Path(link).name[:-18]
            if youtube_id in self.__invalid_files__:
                continue
            files.append((youtube_id, link))
        return files

#__example__ from rsp.ml.dataset import UCF101
#__example__ import rsp.ml.multi_transforms as multi_transforms
#__example__ import cv2 as cv
#__example__ 
#__example__ transforms = multi_transforms.Compose([
#__example__     multi_transforms.Color(1.5, p=0.5),
#__example__     multi_transforms.Stack()
#__example__ ])
#__example__ ds = UCF101('train', fold=1, transforms=transforms)
#__example__ 
#__example__ for X, T in ds:
#__example__   for x in X.permute(0, 2, 3, 1):
#__example__     img_color = x[:, :, :3].numpy()
#__example__     img_depth = x[:, :, 3].numpy()
#__example__ 
#__example__     cv.imshow('color', img_color)
#__example__     cv.imshow('depth', img_depth)
#__example__ 
#__example__     cv.waitKey(30)
class UCF101(Dataset):
    def __init__(
            self,
            split:str,
            fold:int = None,
            cache_dir:str = None,
            force_reload:bool = False,
            target_size = (400, 400),
            sequence_length:int = 60,
            transforms:multi_transforms.Compose = multi_transforms.Compose([]),
            verbose:bool = True
        ):
        """
        Initializes a new instance.

        Parameters
        ----------
        split : str
            Dataset split [train|val|test]
        fold : int
            Fold number. The dataset is split into 3 folds. If fold is None, all folds will be loaded.
        cache_dir : str, default = None
            Directory to store the downloaded files. If set to `None`, the default cache directory will be used
        force_reload : bool, default = False
            If set to `True`, the dataset will be reloaded
        target_size : (int, int), default = (400, 400)
            Size of the frames. The frames will be resized to this size.
        sequence_length : int, default = 30
            Length of the sequences
        transforms : rsp.ml.multi_transforms.Compose = default = rsp.ml.multi_transforms.Compose([])
            Transformations, that will be applied to each input sequence. See documentation of `rsp.ml.multi_transforms` for more details.
        verbose : bool, default = False
            If set to `True`, the progress will be printed.
        """
        assert split in ['train', 'val'], f'{split} is not a valid split. Please use one of [train, val].'
        assert fold in [None, 1, 2, 3], f'{fold} is not a valid fold. Please use one of [None, 1, 2, 3].'

        self.download_link = f'https://drive.google.com/uc?id=1AgvxQl9ShkQyh83FGf-njBiwPKkrLkWw'
        self.split = split
        self.fold = fold
        self.force_reload = force_reload
        self.target_size = target_size
        self.sequence_length = sequence_length
        self.transforms = transforms
        self.verbose = verbose

        if cache_dir is None:
            self.__cache_dir__ = Path(user_cache_dir("rsp-ml", "Robert Schulz")).joinpath('dataset', 'UCF101')
        else:
            self.__cache_dir__ = Path(cache_dir).joinpath('UCF101')
        self.__cache_dir__.mkdir(parents=True, exist_ok=True)

        self.__download__()
        self.__list_files__()

    def __len__(self):
        return len(self.__files__)

    def __getitem__(self, index):
        action, fname = self.__files__[index]

        cap = cv.VideoCapture(fname)
        cnt = int(cap.get(cv.CAP_PROP_FRAME_COUNT))

        if cnt-self.sequence_length <= 0:
            start_idx = 0
        else:
            start_idx = np.random.randint(0, cnt-self.sequence_length)
        end_idx = start_idx + self.sequence_length

        frames = []
        cap.set(cv.CAP_PROP_POS_FRAMES, start_idx)
        while True:
            ret, frame = cap.read()
            if not ret:
                break
            frame = cv.resize(frame, self.target_size)
            frames.append(frame)
            if len(frames) >= self.sequence_length:
                break

        X = torch.tensor(np.array(frames), dtype=torch.float32).permute(0, 3, 1, 2) / 255
        T = torch.zeros((len(self.action_labels)), dtype=torch.float32)
        T[action] = 1

        if X.shape[0] < self.sequence_length:
            if self.verbose:
                try:
                    console.warn(f'Seuqnce length was {X.shape[0]}. Expected {self.sequence_length}. Automatic expanding...')
                except:
                    print(f'Seuqnce length was {X.shape[0]}. Expected {self.sequence_length}. Automatic expanding...')
            X = torch.concat([X, torch.zeros((self.sequence_length-X.shape[0], X.shape[1], X.shape[2], X.shape[3]), dtype=torch.float32)])

        X = self.transforms(X)

        return X, T
    
    def __download__(self):
        zip_file = f'{self.__cache_dir__.parent}/UCF101.zip'
        if not os.path.isdir(f'{self.__cache_dir__}') or len(os.listdir(self.__cache_dir__)) == 0 or self.force_reload:
            if self.verbose:
                try:
                    console.print_c('Downloading UCF101 dataset...', color=console.color.GREEN)
                except:
                    print('Downloading UCF101 dataset...')
            if not os.path.isfile(zip_file):
                gdown.download(self.download_link, zip_file, quiet=not self.verbose)
            with zipfile.ZipFile(zip_file, 'r') as zip_ref:
                zip_ref.extractall(self.__cache_dir__.parent)
            if os.path.isdir(f'{self.__cache_dir__.parent}/__MACOSX'):
                shutil.rmtree(f'{self.__cache_dir__.parent}/__MACOSX')
            os.remove(zip_file)

    def __list_files__(self):
        self.__files__ = []

        self.action_labels = sorted([Path(folder).name for folder in glob(f'{self.__cache_dir__}/sequences/*')])

        splitname = 'trainlist' if self.split == 'train' else 'testlist'
        if self.fold is None:
            split_files = glob(f'{self.__cache_dir__}/splits/{splitname}*.txt')
        else:
            split_files = glob(f'{self.__cache_dir__}/splits/{splitname}*{self.fold}.txt')

        for split_file in split_files:
            split_file = Path(split_file)

            with open(split_file, 'r') as file:
                lines = [line for line in file.read().split('\n') if len(line) > 0]
                for line in lines:
                    action = line.split(' ')[0].split('/')[0]
                    fname = line.split(' ')[0]
                    fname = f'{self.__cache_dir__}/sequences/{fname}'
                    action_idx = self.action_labels.index(action)
                    self.__files__.append((action_idx, fname))
                pass
            pass

#__example__ from rsp.ml.dataset import UTKinectAction3D
#__example__ import rsp.ml.multi_transforms as multi_transforms
#__example__ import cv2 as cv
#__example__ 
#__example__ transforms = multi_transforms.Compose([
#__example__     multi_transforms.Color(1.5, p=0.5),
#__example__     multi_transforms.Stack()
#__example__ ])
#__example__ ds = UTKinectAction3D('train', transforms=transforms)
#__example__ 
#__example__ for X, T in ds:
#__example__   for x in X.permute(0, 2, 3, 1):
#__example__     img_color = x[:, :, :3].numpy()
#__example__     img_depth = x[:, :, 3].numpy()
#__example__ 
#__example__     cv.imshow('color', img_color)
#__example__     cv.imshow('depth', img_depth)
#__example__ 
#__example__     cv.waitKey(30)
class UTKinectAction3D(Dataset):
    """
    Dataset class for the UTKinectAction3D dataset.

    Parameters
    ----------
    split : str
        Dataset split [train|val]
    cache_dir : str, default = None
        Directory to store the downloaded files. If set to `None`, the default cache directory will be used
    force_reload : bool, default = False
        If set to `True`, the dataset will be reloaded
    target_size : (int, int), default = (400, 400)
        Size of the frames. The frames will be resized to this size.
    sequence_length : int, default = 30
        Length of the sequences
    transforms : rsp.ml.multi_transforms.Compose = default = rsp.ml.multi_transforms.Compose([])
        Transformations, that will be applied to each input sequence. See documentation of `rsp.ml.multi_transforms` for more details.
    verbose : bool, default = False
        If set to `True`, the progress will be printed.
    """
    def __init__(
            self,
            split:str,
            cache_dir:str = None,
            force_reload:bool = False,
            target_size:tuple[int, int] = (400, 400),
            sequence_length:int = 30,
            transforms:multi_transforms.Compose = multi_transforms.Compose([]),
            verbose:bool = True
        ):
        assert split in ['train', 'val'], f'{split} is not a valid split. Please use one of [train, val].'

        self.downloadlink_rgb = 'https://cvrc.ece.utexas.edu/KinectDatasets/RGB.zip'
        self.downloadlink_depth = 'https://cvrc.ece.utexas.edu/KinectDatasets/depth.zip'
        self.downloadlink_labels = 'https://cvrc.ece.utexas.edu/KinectDatasets/actionLabel.txt'

        self.split = split
        self.force_reload = force_reload
        self.target_size = target_size
        self.sequence_length = sequence_length
        self.transforms = transforms
        self.verbose = verbose

        if cache_dir is None:
            self.__cache_dir__ = Path(user_cache_dir("rsp-ml", "Robert Schulz")).joinpath('dataset', 'UTKinectAction3D')
        else:
            self.__cache_dir__ = Path(cache_dir).joinpath('UTKinectAction3D')
        self.__cache_dir__.mkdir(parents=True, exist_ok=True)

        self.__download__()
        self.__list_files__()

    def __download__(self):
        # RGB
        zip_file_rgb = f'{self.__cache_dir__.parent}/UTKinectAction3D_rgb.zip'
        if not self.__cache_dir__.joinpath('RGB').exists() or\
                len(os.listdir(self.__cache_dir__.joinpath('RGB'))) == 0 or\
                self.force_reload:
            if not os.path.isfile(zip_file_rgb):
                print(f'Downloading {self.downloadlink_rgb}')
                urllib.request.urlretrieve(self.downloadlink_rgb, zip_file_rgb)
            zipfile.ZipFile(zip_file_rgb, 'r').extractall(self.__cache_dir__)
        if os.path.isfile(zip_file_rgb):
            os.remove(zip_file_rgb)

        # depth
        zip_file_depth = f'{self.__cache_dir__.parent}/UTKinectAction3D_depth.zip'
        if not self.__cache_dir__.joinpath('depth').exists() or\
                len(os.listdir(self.__cache_dir__.joinpath('depth'))) == 0 or\
                self.force_reload:
            if not os.path.isfile(zip_file_depth):
                print(f'Downloading {self.downloadlink_depth}')
                urllib.request.urlretrieve(self.downloadlink_depth, zip_file_depth)
            zipfile.ZipFile(zip_file_depth, 'r').extractall(self.__cache_dir__)
        if os.path.isfile(zip_file_depth):
            os.remove(zip_file_depth)

        # labels
        labels_txt = f'{self.__cache_dir__}/labels.txt'
        if not os.path.isfile(labels_txt) or self.force_reload:
            print(f'Downloading {self.downloadlink_labels}')
            urllib.request.urlretrieve(self.downloadlink_labels, labels_txt)

    def __list_files__(self):
        self.__sequences__ = []
        self.action_labels = []

        labels_txt = f'{self.__cache_dir__}/labels.txt'
        with open(labels_txt, 'r') as file:
            lines = [line for line in file.read().split('\n') if len(line) > 0]
        
        sub_dir = None
        for line in lines:
            line = line.replace('  ', ' ')
            if ':' in line:
                try:
                    subject = int(sub_dir[1:3])
                    if self.split == 'train' and subject <= 8 or\
                        self.split == 'val' and subject > 8:
                        action = line.split(':')[0]
                        if action not in self.action_labels:
                            self.action_labels.append(action)
                        start_idx = int(line.split(' ')[1])
                        end_idx = int(line.split(' ')[2])
                        self.__sequences__.append((sub_dir, action, start_idx, end_idx))
                except Exception as e:
                    pass
            else:
                sub_dir = line
        self.action_labels = sorted(self.action_labels)

    def __len__(self):
        return len(self.__sequences__)
    
    def __getitem__(self, index):
        sub_dir, action_label, start_idx, end_idx = self.__sequences__[index]
        action_idx = self.action_labels.index(action_label)
        
        def load_utkinect_depth_from_xml(file_path):
            """ Lädt eine UTKinect XML-Tiefendatei und konvertiert sie in ein 320x240 NumPy-Array """
            
            # XML-Datei parsen
            tree = ET.parse(file_path)
            root = tree.getroot()

            # Höhe und Breite aus XML auslesen
            width = int(root.find(".//width").text)
            height = int(root.find(".//height").text)

            # Tiefendaten aus <data>-Tag holen (Werte sind durch Leerzeichen getrennt)
            depth_text = root.find(".//data").text.strip()
            depth_values = np.array(list(map(int, depth_text.split())), dtype=np.uint16)

            # In (Höhe, Breite) umformen
            depth_image = depth_values.reshape((height, width))

            return depth_image

        rgb_frames, depth_frames = [], []
        for i in range(start_idx, end_idx+1):
            rgb_file = self.__cache_dir__.joinpath('RGB', sub_dir, f'colorImg{i}.jpg')
            if rgb_file.exists() == False:
                continue
            rgb_img = cv.imread(str(rgb_file))
            rgb_img = cv.resize(rgb_img, (self.target_size[0], self.target_size[1]))

            depth_file = self.__cache_dir__.joinpath('depth', sub_dir, f'depthImg{i}.xml')
            depth_img = load_utkinect_depth_from_xml(depth_file)
            depth_img = cv.resize(depth_img, (self.target_size[0], self.target_size[1]))

            rgb_frames.append(rgb_img)
            depth_frames.append(depth_img)

        X_rgb = torch.tensor(np.array(rgb_frames), dtype=torch.float32).permute(0, 3, 1, 2) / 255
        X_depth = torch.tensor(np.array(depth_frames), dtype=torch.float32).unsqueeze(1) / 31800
        X = torch.cat([X_rgb, X_depth], dim=1)

        if X.shape[0] > self.sequence_length:
            start_idx = np.random.randint(0, X.shape[0]-self.sequence_length)
            X = X[start_idx:start_idx+self.sequence_length]

        X = self.transforms(X)

        T = torch.zeros((len(self.action_labels)), dtype=torch.float32)
        T[action_idx] = 1

        return X, T

if __name__ == '__main__':
    USE_DEPTH_DATA = True
    backgrounds = TUCRID.load_backgrounds(USE_DEPTH_DATA)
    tranforms_train = multi_transforms.Compose([
        multi_transforms.ReplaceBackground(
            backgrounds = backgrounds,
            hsv_filter=[(69, 87, 139, 255, 52, 255)],
            p = 0.8,
            rotate=180
        ),
        multi_transforms.Resize((400, 400), auto_crop=False),
        multi_transforms.Color(0.1, p = 0.2),
        multi_transforms.Brightness(0.7, 1.3),
        multi_transforms.Satturation(0.7, 1.3),
        multi_transforms.RandomHorizontalFlip(),
        multi_transforms.GaussianNoise(0.002),
        multi_transforms.Rotate(max_angle=3),
        multi_transforms.Stack()
    ])
    tucrid_train = TUCRID('train', load_depth_data=USE_DEPTH_DATA, transforms=tranforms_train)
    sampler_train = tucrid_train.get_uniform_sampler()

    dl_train = DataLoader(tucrid_train, batch_size=4, sampler=sampler_train)

    for X, T in dl_train:
        for seq_X, seq_T in zip(X, T):
            for x in seq_X:
                img = x[0:3].permute(1, 2, 0).numpy()
                img = np.array(img * 255, dtype=np.uint8)

                if x.shape[0] == 4:
                    img_depth = x[3].numpy()
                    img_depth = cv.applyColorMap((img_depth * 255).astype(np.uint8), cv.COLORMAP_COOL)
                    img = np.hstack([img, img_depth])

                cv.imshow('img', img)
                cv.waitKey(30)
    pass


    k400 = Kinetics('train', num_threads=2, cache_dir='/Volumes/USB-Freigabe/KINETICS400')#cache_dir='/Volumes/ROBERT512GB/KINETICS400')

    for i, (X, T) in enumerate(k400):
        print(i)
        pass


import torch
import torch.nn as nn
import torch.nn.functional as F

class SelfAttention(nn.Module):
    def __init__(
            self,
            d_q:int = 2,
            d_k:int = 2,
            d_v:int = 4,
            embed_dim:int = 3
        ):
        super().__init__()

        self.d_q = d_q
        self.d_k = d_k
        self.d_v = d_v

        self.W_q = nn.Parameter(torch.rand(embed_dim, d_q))
        self.W_k = nn.Parameter(torch.rand(embed_dim, d_k))
        self.W_v = nn.Parameter(torch.rand(embed_dim, d_v))
        pass

    def forward(self, X):
        Z = []
        # iterate over batch_size
        for x in X:
            Q = x @ self.W_q    # Queries
            K = x @ self.W_k    # Keys
            V = x @ self.W_v    # Values

            omega = Q @ K.T                                     # omega ...unnormalized attantion weights
            alpha = F.softmax(omega / self.d_k**0.5, dim=0)     # alpha ...normalized attention weights
            z = alpha @ V                                       # z     ...context vector -> attention-weighted version of original query input x_i
            Z.append(z)
        
        Z = torch.stack(Z)
        return Z

class MultiHeadSelfAttention(nn.Module):
    def __init__(
            self,
            num_heads:int,
            d_q:int = 2,
            d_k:int = 2,
            d_v:int = 4,
            embed_dim:int = 3
        ):
        super().__init__()

        self.d_q = d_q
        self.d_k = d_k
        self.d_v = d_v

        self.heads = nn.ModuleList([SelfAttention(d_q, d_k, d_v, embed_dim) for _ in range(num_heads)])

    def forward(self, X):
        return torch.cat([head(X) for head in self.heads], dim=-1)
